/****************************************************************************
** Meta object code from reading C++ file 'qquick3dspatialsound_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/spatialaudioquick3d/qquick3dspatialsound_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qquick3dspatialsound_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN20QQuick3DSpatialSoundE_t {};
} // unnamed namespace

template <> constexpr inline auto QQuick3DSpatialSound::qt_create_metaobjectdata<qt_meta_tag_ZN20QQuick3DSpatialSoundE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QQuick3DSpatialSound",
        "QML.Element",
        "SpatialSound",
        "sourceChanged",
        "",
        "volumeChanged",
        "distanceModelChanged",
        "sizeChanged",
        "distanceCutoffChanged",
        "manualAttenuationChanged",
        "occlusionIntensityChanged",
        "directivityChanged",
        "directivityOrderChanged",
        "nearFieldGainChanged",
        "loopsChanged",
        "autoPlayChanged",
        "play",
        "pause",
        "stop",
        "updatePosition",
        "updateRotation",
        "source",
        "QUrl",
        "volume",
        "distanceModel",
        "DistanceModel",
        "size",
        "distanceCutoff",
        "manualAttenuation",
        "occlusionIntensity",
        "directivity",
        "directivityOrder",
        "nearFieldGain",
        "loops",
        "autoPlay",
        "Logarithmic",
        "Linear",
        "ManualAttenuation",
        "Loops",
        "Infinite",
        "Once"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sourceChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'volumeChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'distanceModelChanged'
        QtMocHelpers::SignalData<void()>(6, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'sizeChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'distanceCutoffChanged'
        QtMocHelpers::SignalData<void()>(8, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'manualAttenuationChanged'
        QtMocHelpers::SignalData<void()>(9, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'occlusionIntensityChanged'
        QtMocHelpers::SignalData<void()>(10, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'directivityChanged'
        QtMocHelpers::SignalData<void()>(11, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'directivityOrderChanged'
        QtMocHelpers::SignalData<void()>(12, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'nearFieldGainChanged'
        QtMocHelpers::SignalData<void()>(13, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'loopsChanged'
        QtMocHelpers::SignalData<void()>(14, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'autoPlayChanged'
        QtMocHelpers::SignalData<void()>(15, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(16, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'pause'
        QtMocHelpers::SlotData<void()>(17, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(18, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'updatePosition'
        QtMocHelpers::SlotData<void()>(19, 4, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'updateRotation'
        QtMocHelpers::SlotData<void()>(20, 4, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'source'
        QtMocHelpers::PropertyData<QUrl>(21, 0x80000000 | 22, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'volume'
        QtMocHelpers::PropertyData<float>(23, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 1),
        // property 'distanceModel'
        QtMocHelpers::PropertyData<enum DistanceModel>(24, 0x80000000 | 25, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 2),
        // property 'size'
        QtMocHelpers::PropertyData<float>(26, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3),
        // property 'distanceCutoff'
        QtMocHelpers::PropertyData<float>(27, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'manualAttenuation'
        QtMocHelpers::PropertyData<float>(28, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'occlusionIntensity'
        QtMocHelpers::PropertyData<float>(29, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 6),
        // property 'directivity'
        QtMocHelpers::PropertyData<float>(30, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 7),
        // property 'directivityOrder'
        QtMocHelpers::PropertyData<float>(31, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 8),
        // property 'nearFieldGain'
        QtMocHelpers::PropertyData<float>(32, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 9),
        // property 'loops'
        QtMocHelpers::PropertyData<int>(33, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 10),
        // property 'autoPlay'
        QtMocHelpers::PropertyData<bool>(34, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 11),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'DistanceModel'
        QtMocHelpers::EnumData<enum DistanceModel>(25, 25, QMC::EnumFlags{}).add({
            {   35, DistanceModel::Logarithmic },
            {   36, DistanceModel::Linear },
            {   37, DistanceModel::ManualAttenuation },
        }),
        // enum 'Loops'
        QtMocHelpers::EnumData<enum Loops>(38, 38, QMC::EnumFlags{}).add({
            {   39, Loops::Infinite },
            {   40, Loops::Once },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QQuick3DSpatialSound, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QQuick3DSpatialSound::staticMetaObject = { {
    QMetaObject::SuperData::link<QQuick3DNode::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QQuick3DSpatialSoundE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QQuick3DSpatialSoundE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN20QQuick3DSpatialSoundE_t>.metaTypes,
    nullptr
} };

void QQuick3DSpatialSound::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QQuick3DSpatialSound *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sourceChanged(); break;
        case 1: _t->volumeChanged(); break;
        case 2: _t->distanceModelChanged(); break;
        case 3: _t->sizeChanged(); break;
        case 4: _t->distanceCutoffChanged(); break;
        case 5: _t->manualAttenuationChanged(); break;
        case 6: _t->occlusionIntensityChanged(); break;
        case 7: _t->directivityChanged(); break;
        case 8: _t->directivityOrderChanged(); break;
        case 9: _t->nearFieldGainChanged(); break;
        case 10: _t->loopsChanged(); break;
        case 11: _t->autoPlayChanged(); break;
        case 12: _t->play(); break;
        case 13: _t->pause(); break;
        case 14: _t->stop(); break;
        case 15: _t->updatePosition(); break;
        case 16: _t->updateRotation(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::sourceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::volumeChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::distanceModelChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::sizeChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::distanceCutoffChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::manualAttenuationChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::occlusionIntensityChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::directivityChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::directivityOrderChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::nearFieldGainChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::loopsChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DSpatialSound::*)()>(_a, &QQuick3DSpatialSound::autoPlayChanged, 11))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QUrl*>(_v) = _t->source(); break;
        case 1: *reinterpret_cast<float*>(_v) = _t->volume(); break;
        case 2: *reinterpret_cast<enum DistanceModel*>(_v) = _t->distanceModel(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->size(); break;
        case 4: *reinterpret_cast<float*>(_v) = _t->distanceCutoff(); break;
        case 5: *reinterpret_cast<float*>(_v) = _t->manualAttenuation(); break;
        case 6: *reinterpret_cast<float*>(_v) = _t->occlusionIntensity(); break;
        case 7: *reinterpret_cast<float*>(_v) = _t->directivity(); break;
        case 8: *reinterpret_cast<float*>(_v) = _t->directivityOrder(); break;
        case 9: *reinterpret_cast<float*>(_v) = _t->nearFieldGain(); break;
        case 10: *reinterpret_cast<int*>(_v) = _t->loops(); break;
        case 11: *reinterpret_cast<bool*>(_v) = _t->autoPlay(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setSource(*reinterpret_cast<QUrl*>(_v)); break;
        case 1: _t->setVolume(*reinterpret_cast<float*>(_v)); break;
        case 2: _t->setDistanceModel(*reinterpret_cast<enum DistanceModel*>(_v)); break;
        case 3: _t->setSize(*reinterpret_cast<float*>(_v)); break;
        case 4: _t->setDistanceCutoff(*reinterpret_cast<float*>(_v)); break;
        case 5: _t->setManualAttenuation(*reinterpret_cast<float*>(_v)); break;
        case 6: _t->setOcclusionIntensity(*reinterpret_cast<float*>(_v)); break;
        case 7: _t->setDirectivity(*reinterpret_cast<float*>(_v)); break;
        case 8: _t->setDirectivityOrder(*reinterpret_cast<float*>(_v)); break;
        case 9: _t->setNearFieldGain(*reinterpret_cast<float*>(_v)); break;
        case 10: _t->setLoops(*reinterpret_cast<int*>(_v)); break;
        case 11: _t->setAutoPlay(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QQuick3DSpatialSound::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuick3DSpatialSound::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QQuick3DSpatialSoundE_t>.strings))
        return static_cast<void*>(this);
    return QQuick3DNode::qt_metacast(_clname);
}

int QQuick3DSpatialSound::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QQuick3DNode::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 17)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 17;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 17)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 17;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 12;
    }
    return _id;
}

// SIGNAL 0
void QQuick3DSpatialSound::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QQuick3DSpatialSound::volumeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QQuick3DSpatialSound::distanceModelChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QQuick3DSpatialSound::sizeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QQuick3DSpatialSound::distanceCutoffChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QQuick3DSpatialSound::manualAttenuationChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QQuick3DSpatialSound::occlusionIntensityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QQuick3DSpatialSound::directivityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void QQuick3DSpatialSound::directivityOrderChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void QQuick3DSpatialSound::nearFieldGainChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void QQuick3DSpatialSound::loopsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 10, nullptr);
}

// SIGNAL 11
void QQuick3DSpatialSound::autoPlayChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}
QT_WARNING_POP
