// Copyright (C) 2022 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#include "qplatformmediaplugin_p.h"

QT_BEGIN_NAMESPACE

QPlatformMediaPlugin::QPlatformMediaPlugin(QObject *parent) : QObject(parent) { }

QPlatformMediaPlugin::~QPlatformMediaPlugin() = default;

QT_END_NAMESPACE

#include "moc_qplatformmediaplugin_p.cpp"
