/****************************************************************************
** Meta object code from reading C++ file 'qmediacapturesession.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/recording/qmediacapturesession.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qmediacapturesession.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN20QMediaCaptureSessionE_t {};
} // unnamed namespace

template <> constexpr inline auto QMediaCaptureSession::qt_create_metaobjectdata<qt_meta_tag_ZN20QMediaCaptureSessionE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QMediaCaptureSession",
        "audioInputChanged",
        "",
        "audioBufferInputChanged",
        "cameraChanged",
        "nativeVideoSourceChanged",
        "screenCaptureChanged",
        "windowCaptureChanged",
        "videoFrameInputChanged",
        "imageCaptureChanged",
        "recorderChanged",
        "videoOutputChanged",
        "audioOutputChanged",
        "audioInput",
        "QAudioInput*",
        "audioBufferInput",
        "QAudioBufferInput*",
        "audioOutput",
        "QAudioOutput*",
        "camera",
        "QCamera*",
        "nativeVideoSource",
        "screenCapture",
        "QScreenCapture*",
        "windowCapture",
        "QWindowCapture*",
        "videoFrameInput",
        "QVideoFrameInput*",
        "imageCapture",
        "QImageCapture*",
        "recorder",
        "QMediaRecorder*",
        "videoOutput"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'audioInputChanged'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioBufferInputChanged'
        QtMocHelpers::RevisionedSignalData<void()>(3, 2, QMC::AccessPublic, 0x608, QMetaType::Void),
        // Signal 'cameraChanged'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'nativeVideoSourceChanged'
        QtMocHelpers::RevisionedSignalData<void()>(5, 2, QMC::AccessPublic, 0x60c, QMetaType::Void),
        // Signal 'screenCaptureChanged'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'windowCaptureChanged'
        QtMocHelpers::SignalData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoFrameInputChanged'
        QtMocHelpers::RevisionedSignalData<void()>(8, 2, QMC::AccessPublic, 0x608, QMetaType::Void),
        // Signal 'imageCaptureChanged'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'recorderChanged'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoOutputChanged'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioOutputChanged'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'audioInput'
        QtMocHelpers::PropertyData<QAudioInput*>(13, 0x80000000 | 14, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'audioBufferInput'
        QtMocHelpers::PropertyData<QAudioBufferInput*>(15, 0x80000000 | 16, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1, 0x608),
        // property 'audioOutput'
        QtMocHelpers::PropertyData<QAudioOutput*>(17, 0x80000000 | 18, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 10),
        // property 'camera'
        QtMocHelpers::PropertyData<QCamera*>(19, 0x80000000 | 20, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 2),
        // property 'nativeVideoSource'
        QtMocHelpers::PropertyData<QObject*>(21, QMetaType::QObjectStar, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3, 0x60c),
        // property 'screenCapture'
        QtMocHelpers::PropertyData<QScreenCapture*>(22, 0x80000000 | 23, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 4),
        // property 'windowCapture'
        QtMocHelpers::PropertyData<QWindowCapture*>(24, 0x80000000 | 25, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 5),
        // property 'videoFrameInput'
        QtMocHelpers::PropertyData<QVideoFrameInput*>(26, 0x80000000 | 27, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 6, 0x608),
        // property 'imageCapture'
        QtMocHelpers::PropertyData<QImageCapture*>(28, 0x80000000 | 29, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 7),
        // property 'recorder'
        QtMocHelpers::PropertyData<QMediaRecorder*>(30, 0x80000000 | 31, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 8),
        // property 'videoOutput'
        QtMocHelpers::PropertyData<QObject*>(32, QMetaType::QObjectStar, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 9),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QMediaCaptureSession, qt_meta_tag_ZN20QMediaCaptureSessionE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QMediaCaptureSession::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QMediaCaptureSessionE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QMediaCaptureSessionE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN20QMediaCaptureSessionE_t>.metaTypes,
    nullptr
} };

void QMediaCaptureSession::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QMediaCaptureSession *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->audioInputChanged(); break;
        case 1: _t->audioBufferInputChanged(); break;
        case 2: _t->cameraChanged(); break;
        case 3: _t->nativeVideoSourceChanged(); break;
        case 4: _t->screenCaptureChanged(); break;
        case 5: _t->windowCaptureChanged(); break;
        case 6: _t->videoFrameInputChanged(); break;
        case 7: _t->imageCaptureChanged(); break;
        case 8: _t->recorderChanged(); break;
        case 9: _t->videoOutputChanged(); break;
        case 10: _t->audioOutputChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::audioInputChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::audioBufferInputChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::cameraChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::nativeVideoSourceChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::screenCaptureChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::windowCaptureChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::videoFrameInputChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::imageCaptureChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::recorderChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::videoOutputChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaCaptureSession::*)()>(_a, &QMediaCaptureSession::audioOutputChanged, 10))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QAudioInput**>(_v) = _t->audioInput(); break;
        case 1: *reinterpret_cast<QAudioBufferInput**>(_v) = _t->audioBufferInput(); break;
        case 2: *reinterpret_cast<QAudioOutput**>(_v) = _t->audioOutput(); break;
        case 3: *reinterpret_cast<QCamera**>(_v) = _t->camera(); break;
        case 4: *reinterpret_cast<QObject**>(_v) = _t->nativeVideoSource(); break;
        case 5: *reinterpret_cast<QScreenCapture**>(_v) = _t->screenCapture(); break;
        case 6: *reinterpret_cast<QWindowCapture**>(_v) = _t->windowCapture(); break;
        case 7: *reinterpret_cast<QVideoFrameInput**>(_v) = _t->videoFrameInput(); break;
        case 8: *reinterpret_cast<QImageCapture**>(_v) = _t->imageCapture(); break;
        case 9: *reinterpret_cast<QMediaRecorder**>(_v) = _t->recorder(); break;
        case 10: *reinterpret_cast<QObject**>(_v) = _t->videoOutput(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setAudioInput(*reinterpret_cast<QAudioInput**>(_v)); break;
        case 1: _t->setAudioBufferInput(*reinterpret_cast<QAudioBufferInput**>(_v)); break;
        case 2: _t->setAudioOutput(*reinterpret_cast<QAudioOutput**>(_v)); break;
        case 3: _t->setCamera(*reinterpret_cast<QCamera**>(_v)); break;
        case 4: _t->setNativeVideoSource(*reinterpret_cast<QObject**>(_v)); break;
        case 5: _t->setScreenCapture(*reinterpret_cast<QScreenCapture**>(_v)); break;
        case 6: _t->setWindowCapture(*reinterpret_cast<QWindowCapture**>(_v)); break;
        case 7: _t->setVideoFrameInput(*reinterpret_cast<QVideoFrameInput**>(_v)); break;
        case 8: _t->setImageCapture(*reinterpret_cast<QImageCapture**>(_v)); break;
        case 9: _t->setRecorder(*reinterpret_cast<QMediaRecorder**>(_v)); break;
        case 10: _t->setVideoOutput(*reinterpret_cast<QObject**>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QMediaCaptureSession::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QMediaCaptureSession::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20QMediaCaptureSessionE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QMediaCaptureSession::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 11;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    return _id;
}

// SIGNAL 0
void QMediaCaptureSession::audioInputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QMediaCaptureSession::audioBufferInputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QMediaCaptureSession::cameraChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QMediaCaptureSession::nativeVideoSourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QMediaCaptureSession::screenCaptureChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QMediaCaptureSession::windowCaptureChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QMediaCaptureSession::videoFrameInputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QMediaCaptureSession::imageCaptureChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void QMediaCaptureSession::recorderChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void QMediaCaptureSession::videoOutputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void QMediaCaptureSession::audioOutputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 10, nullptr);
}
QT_WARNING_POP
