/****************************************************************************
** Meta object code from reading C++ file 'qimagecapture.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/camera/qimagecapture.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qimagecapture.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QImageCaptureE_t {};
} // unnamed namespace

template <> constexpr inline auto QImageCapture::qt_create_metaobjectdata<qt_meta_tag_ZN13QImageCaptureE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QImageCapture",
        "errorChanged",
        "",
        "errorOccurred",
        "id",
        "QImageCapture::Error",
        "error",
        "errorString",
        "readyForCaptureChanged",
        "ready",
        "metaDataChanged",
        "fileFormatChanged",
        "qualityChanged",
        "resolutionChanged",
        "imageExposed",
        "imageCaptured",
        "QImage",
        "preview",
        "imageMetadataAvailable",
        "QMediaMetaData",
        "metaData",
        "imageAvailable",
        "QVideoFrame",
        "frame",
        "imageSaved",
        "fileName",
        "captureToFile",
        "location",
        "capture",
        "_q_error",
        "readyForCapture",
        "Error",
        "fileFormat",
        "FileFormat",
        "quality",
        "Quality",
        "supportedFormats",
        "QList<FileFormat>",
        "NoError",
        "NotReadyError",
        "ResourceError",
        "OutOfSpaceError",
        "NotSupportedFeatureError",
        "FormatError",
        "VeryLowQuality",
        "LowQuality",
        "NormalQuality",
        "HighQuality",
        "VeryHighQuality",
        "UnspecifiedFormat",
        "JPEG",
        "PNG",
        "WebP",
        "Tiff",
        "LastFileFormat"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'errorChanged'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'errorOccurred'
        QtMocHelpers::SignalData<void(int, QImageCapture::Error, const QString &)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 }, { 0x80000000 | 5, 6 }, { QMetaType::QString, 7 },
        }}),
        // Signal 'readyForCaptureChanged'
        QtMocHelpers::SignalData<void(bool)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 9 },
        }}),
        // Signal 'metaDataChanged'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'fileFormatChanged'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'qualityChanged'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'resolutionChanged'
        QtMocHelpers::SignalData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'imageExposed'
        QtMocHelpers::SignalData<void(int)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 },
        }}),
        // Signal 'imageCaptured'
        QtMocHelpers::SignalData<void(int, const QImage &)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 }, { 0x80000000 | 16, 17 },
        }}),
        // Signal 'imageMetadataAvailable'
        QtMocHelpers::SignalData<void(int, const QMediaMetaData &)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 }, { 0x80000000 | 19, 20 },
        }}),
        // Signal 'imageAvailable'
        QtMocHelpers::SignalData<void(int, const QVideoFrame &)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 }, { 0x80000000 | 22, 23 },
        }}),
        // Signal 'imageSaved'
        QtMocHelpers::SignalData<void(int, const QString &)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 }, { QMetaType::QString, 25 },
        }}),
        // Slot 'captureToFile'
        QtMocHelpers::SlotData<int(const QString &)>(26, 2, QMC::AccessPublic, QMetaType::Int, {{
            { QMetaType::QString, 27 },
        }}),
        // Slot 'captureToFile'
        QtMocHelpers::SlotData<int()>(26, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Int),
        // Slot 'capture'
        QtMocHelpers::SlotData<int()>(28, 2, QMC::AccessPublic, QMetaType::Int),
        // Slot '_q_error'
        QtMocHelpers::SlotData<void(int, int, const QString &)>(29, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 }, { QMetaType::Int, 2 }, { QMetaType::QString, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'readyForCapture'
        QtMocHelpers::PropertyData<bool>(30, QMetaType::Bool, QMC::DefaultPropertyFlags, 2),
        // property 'metaData'
        QtMocHelpers::PropertyData<QMediaMetaData>(20, 0x80000000 | 19, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'error'
        QtMocHelpers::PropertyData<enum Error>(6, 0x80000000 | 31, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 0),
        // property 'errorString'
        QtMocHelpers::PropertyData<QString>(7, QMetaType::QString, QMC::DefaultPropertyFlags, 0),
        // property 'fileFormat'
        QtMocHelpers::PropertyData<enum FileFormat>(32, 0x80000000 | 33, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 4),
        // property 'quality'
        QtMocHelpers::PropertyData<enum Quality>(34, 0x80000000 | 35, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 5),
        // property 'supportedFormats'
        QtMocHelpers::PropertyData<QList<FileFormat>>(36, 0x80000000 | 37, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant, -1, 0x60a),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Error'
        QtMocHelpers::EnumData<enum Error>(31, 31, QMC::EnumFlags{}).add({
            {   38, Error::NoError },
            {   39, Error::NotReadyError },
            {   40, Error::ResourceError },
            {   41, Error::OutOfSpaceError },
            {   42, Error::NotSupportedFeatureError },
            {   43, Error::FormatError },
        }),
        // enum 'Quality'
        QtMocHelpers::EnumData<enum Quality>(35, 35, QMC::EnumFlags{}).add({
            {   44, Quality::VeryLowQuality },
            {   45, Quality::LowQuality },
            {   46, Quality::NormalQuality },
            {   47, Quality::HighQuality },
            {   48, Quality::VeryHighQuality },
        }),
        // enum 'FileFormat'
        QtMocHelpers::EnumData<enum FileFormat>(33, 33, QMC::EnumFlags{}).add({
            {   49, FileFormat::UnspecifiedFormat },
            {   50, FileFormat::JPEG },
            {   51, FileFormat::PNG },
            {   52, FileFormat::WebP },
            {   53, FileFormat::Tiff },
            {   54, FileFormat::LastFileFormat },
        }),
    };
    return QtMocHelpers::metaObjectData<QImageCapture, qt_meta_tag_ZN13QImageCaptureE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QImageCapture::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QImageCaptureE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QImageCaptureE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QImageCaptureE_t>.metaTypes,
    nullptr
} };

void QImageCapture::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QImageCapture *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->errorChanged(); break;
        case 1: _t->errorOccurred((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QImageCapture::Error>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 2: _t->readyForCaptureChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 3: _t->metaDataChanged(); break;
        case 4: _t->fileFormatChanged(); break;
        case 5: _t->qualityChanged(); break;
        case 6: _t->resolutionChanged(); break;
        case 7: _t->imageExposed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 8: _t->imageCaptured((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QImage>>(_a[2]))); break;
        case 9: _t->imageMetadataAvailable((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QMediaMetaData>>(_a[2]))); break;
        case 10: _t->imageAvailable((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVideoFrame>>(_a[2]))); break;
        case 11: _t->imageSaved((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 12: { int _r = _t->captureToFile((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<int*>(_a[0]) = std::move(_r); }  break;
        case 13: { int _r = _t->captureToFile();
            if (_a[0]) *reinterpret_cast<int*>(_a[0]) = std::move(_r); }  break;
        case 14: { int _r = _t->capture();
            if (_a[0]) *reinterpret_cast<int*>(_a[0]) = std::move(_r); }  break;
        case 15: _t->d_func()->_q_error((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QVideoFrame >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)()>(_a, &QImageCapture::errorChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(int , QImageCapture::Error , const QString & )>(_a, &QImageCapture::errorOccurred, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(bool )>(_a, &QImageCapture::readyForCaptureChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)()>(_a, &QImageCapture::metaDataChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)()>(_a, &QImageCapture::fileFormatChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)()>(_a, &QImageCapture::qualityChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)()>(_a, &QImageCapture::resolutionChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(int )>(_a, &QImageCapture::imageExposed, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(int , const QImage & )>(_a, &QImageCapture::imageCaptured, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(int , const QMediaMetaData & )>(_a, &QImageCapture::imageMetadataAvailable, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(int , const QVideoFrame & )>(_a, &QImageCapture::imageAvailable, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (QImageCapture::*)(int , const QString & )>(_a, &QImageCapture::imageSaved, 11))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<bool*>(_v) = _t->isReadyForCapture(); break;
        case 1: *reinterpret_cast<QMediaMetaData*>(_v) = _t->metaData(); break;
        case 2: *reinterpret_cast<enum Error*>(_v) = _t->error(); break;
        case 3: *reinterpret_cast<QString*>(_v) = _t->errorString(); break;
        case 4: *reinterpret_cast<enum FileFormat*>(_v) = _t->fileFormat(); break;
        case 5: *reinterpret_cast<enum Quality*>(_v) = _t->quality(); break;
        case 6: *reinterpret_cast<QList<FileFormat>*>(_v) = _t->supportedFormats(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 1: _t->setMetaData(*reinterpret_cast<QMediaMetaData*>(_v)); break;
        case 4: _t->setFileFormat(*reinterpret_cast<enum FileFormat*>(_v)); break;
        case 5: _t->setQuality(*reinterpret_cast<enum Quality*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QImageCapture::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QImageCapture::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QImageCaptureE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QImageCapture::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 16)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 16;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 16)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 16;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void QImageCapture::errorChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QImageCapture::errorOccurred(int _t1, QImageCapture::Error _t2, const QString & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2, _t3);
}

// SIGNAL 2
void QImageCapture::readyForCaptureChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QImageCapture::metaDataChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QImageCapture::fileFormatChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QImageCapture::qualityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QImageCapture::resolutionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QImageCapture::imageExposed(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void QImageCapture::imageCaptured(int _t1, const QImage & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1, _t2);
}

// SIGNAL 9
void QImageCapture::imageMetadataAvailable(int _t1, const QMediaMetaData & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1, _t2);
}

// SIGNAL 10
void QImageCapture::imageAvailable(int _t1, const QVideoFrame & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1, _t2);
}

// SIGNAL 11
void QImageCapture::imageSaved(int _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1, _t2);
}
QT_WARNING_POP
