/****************************************************************************
** Meta object code from reading C++ file 'qcameradevice.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/camera/qcameradevice.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qcameradevice.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QCameraFormatE_t {};
} // unnamed namespace

template <> constexpr inline auto QCameraFormat::qt_create_metaobjectdata<qt_meta_tag_ZN13QCameraFormatE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QCameraFormat",
        "resolution",
        "QSize",
        "pixelFormat",
        "QVideoFrameFormat::PixelFormat",
        "minFrameRate",
        "maxFrameRate"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
        // property 'resolution'
        QtMocHelpers::PropertyData<QSize>(1, 0x80000000 | 2, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'pixelFormat'
        QtMocHelpers::PropertyData<QVideoFrameFormat::PixelFormat>(3, 0x80000000 | 4, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'minFrameRate'
        QtMocHelpers::PropertyData<float>(5, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'maxFrameRate'
        QtMocHelpers::PropertyData<float>(6, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Constant),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QCameraFormat, qt_meta_tag_ZN13QCameraFormatE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT static const QMetaObject::SuperData qt_meta_extradata_ZN13QCameraFormatE[] = {
    QMetaObject::SuperData::link<QVideoFrameFormat::staticMetaObject>(),
    nullptr
};

Q_CONSTINIT const QMetaObject QCameraFormat::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCameraFormatE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCameraFormatE_t>.data,
    qt_static_metacall,
    qt_meta_extradata_ZN13QCameraFormatE,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QCameraFormatE_t>.metaTypes,
    nullptr
} };

void QCameraFormat::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = reinterpret_cast<QCameraFormat *>(_o);
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QSize*>(_v) = _t->resolution(); break;
        case 1: *reinterpret_cast<QVideoFrameFormat::PixelFormat*>(_v) = _t->pixelFormat(); break;
        case 2: *reinterpret_cast<float*>(_v) = _t->minFrameRate(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->maxFrameRate(); break;
        default: break;
        }
    }
}
namespace {
struct qt_meta_tag_ZN13QCameraDeviceE_t {};
} // unnamed namespace

template <> constexpr inline auto QCameraDevice::qt_create_metaobjectdata<qt_meta_tag_ZN13QCameraDeviceE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QCameraDevice",
        "id",
        "description",
        "isDefault",
        "position",
        "Position",
        "videoFormats",
        "QList<QCameraFormat>",
        "correctionAngle",
        "QtVideo::Rotation",
        "UnspecifiedPosition",
        "BackFace",
        "FrontFace"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
        // property 'id'
        QtMocHelpers::PropertyData<QByteArray>(1, QMetaType::QByteArray, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'description'
        QtMocHelpers::PropertyData<QString>(2, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'isDefault'
        QtMocHelpers::PropertyData<bool>(3, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'position'
        QtMocHelpers::PropertyData<enum Position>(4, 0x80000000 | 5, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'videoFormats'
        QtMocHelpers::PropertyData<QList<QCameraFormat>>(6, 0x80000000 | 7, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'correctionAngle'
        QtMocHelpers::PropertyData<QtVideo::Rotation>(8, 0x80000000 | 9, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Position'
        QtMocHelpers::EnumData<enum Position>(5, 5, QMC::EnumFlags{}).add({
            {   10, Position::UnspecifiedPosition },
            {   11, Position::BackFace },
            {   12, Position::FrontFace },
        }),
    };
    return QtMocHelpers::metaObjectData<QCameraDevice, qt_meta_tag_ZN13QCameraDeviceE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT static const QMetaObject::SuperData qt_meta_extradata_ZN13QCameraDeviceE[] = {
    QMetaObject::SuperData::link<QtVideo::staticMetaObject>(),
    nullptr
};

Q_CONSTINIT const QMetaObject QCameraDevice::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCameraDeviceE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCameraDeviceE_t>.data,
    qt_static_metacall,
    qt_meta_extradata_ZN13QCameraDeviceE,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QCameraDeviceE_t>.metaTypes,
    nullptr
} };

void QCameraDevice::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = reinterpret_cast<QCameraDevice *>(_o);
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QByteArray*>(_v) = _t->id(); break;
        case 1: *reinterpret_cast<QString*>(_v) = _t->description(); break;
        case 2: *reinterpret_cast<bool*>(_v) = _t->isDefault(); break;
        case 3: *reinterpret_cast<enum Position*>(_v) = _t->position(); break;
        case 4: *reinterpret_cast<QList<QCameraFormat>*>(_v) = _t->videoFormats(); break;
        case 5: *reinterpret_cast<QtVideo::Rotation*>(_v) = _t->correctionAngle(); break;
        default: break;
        }
    }
}
QT_WARNING_POP
