/****************************************************************************
** Meta object code from reading C++ file 'qvideoframeformat.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/video/qvideoframeformat.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qvideoframeformat.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN17QVideoFrameFormatE_t {};
} // unnamed namespace

template <> constexpr inline auto QVideoFrameFormat::qt_create_metaobjectdata<qt_meta_tag_ZN17QVideoFrameFormatE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QVideoFrameFormat",
        "PixelFormat",
        "Format_Invalid",
        "Format_ARGB8888",
        "Format_ARGB8888_Premultiplied",
        "Format_XRGB8888",
        "Format_BGRA8888",
        "Format_BGRA8888_Premultiplied",
        "Format_BGRX8888",
        "Format_ABGR8888",
        "Format_XBGR8888",
        "Format_RGBA8888",
        "Format_RGBX8888",
        "Format_AYUV",
        "Format_AYUV_Premultiplied",
        "Format_YUV420P",
        "Format_YUV422P",
        "Format_YV12",
        "Format_UYVY",
        "Format_YUYV",
        "Format_NV12",
        "Format_NV21",
        "Format_IMC1",
        "Format_IMC2",
        "Format_IMC3",
        "Format_IMC4",
        "Format_Y8",
        "Format_Y16",
        "Format_P010",
        "Format_P016",
        "Format_SamplerExternalOES",
        "Format_Jpeg",
        "Format_SamplerRect",
        "Format_YUV420P10",
        "Direction",
        "TopToBottom",
        "BottomToTop",
        "YCbCrColorSpace",
        "YCbCr_Undefined",
        "YCbCr_BT601",
        "YCbCr_BT709",
        "YCbCr_xvYCC601",
        "YCbCr_xvYCC709",
        "YCbCr_JPEG",
        "YCbCr_BT2020",
        "ColorSpace",
        "ColorSpace_Undefined",
        "ColorSpace_BT601",
        "ColorSpace_BT709",
        "ColorSpace_AdobeRgb",
        "ColorSpace_BT2020",
        "ColorTransfer",
        "ColorTransfer_Unknown",
        "ColorTransfer_BT709",
        "ColorTransfer_BT601",
        "ColorTransfer_Linear",
        "ColorTransfer_Gamma22",
        "ColorTransfer_Gamma28",
        "ColorTransfer_ST2084",
        "ColorTransfer_STD_B67",
        "ColorRange",
        "ColorRange_Unknown",
        "ColorRange_Video",
        "ColorRange_Full"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'PixelFormat'
        QtMocHelpers::EnumData<enum PixelFormat>(1, 1, QMC::EnumFlags{}).add({
            {    2, PixelFormat::Format_Invalid },
            {    3, PixelFormat::Format_ARGB8888 },
            {    4, PixelFormat::Format_ARGB8888_Premultiplied },
            {    5, PixelFormat::Format_XRGB8888 },
            {    6, PixelFormat::Format_BGRA8888 },
            {    7, PixelFormat::Format_BGRA8888_Premultiplied },
            {    8, PixelFormat::Format_BGRX8888 },
            {    9, PixelFormat::Format_ABGR8888 },
            {   10, PixelFormat::Format_XBGR8888 },
            {   11, PixelFormat::Format_RGBA8888 },
            {   12, PixelFormat::Format_RGBX8888 },
            {   13, PixelFormat::Format_AYUV },
            {   14, PixelFormat::Format_AYUV_Premultiplied },
            {   15, PixelFormat::Format_YUV420P },
            {   16, PixelFormat::Format_YUV422P },
            {   17, PixelFormat::Format_YV12 },
            {   18, PixelFormat::Format_UYVY },
            {   19, PixelFormat::Format_YUYV },
            {   20, PixelFormat::Format_NV12 },
            {   21, PixelFormat::Format_NV21 },
            {   22, PixelFormat::Format_IMC1 },
            {   23, PixelFormat::Format_IMC2 },
            {   24, PixelFormat::Format_IMC3 },
            {   25, PixelFormat::Format_IMC4 },
            {   26, PixelFormat::Format_Y8 },
            {   27, PixelFormat::Format_Y16 },
            {   28, PixelFormat::Format_P010 },
            {   29, PixelFormat::Format_P016 },
            {   30, PixelFormat::Format_SamplerExternalOES },
            {   31, PixelFormat::Format_Jpeg },
            {   32, PixelFormat::Format_SamplerRect },
            {   33, PixelFormat::Format_YUV420P10 },
        }),
        // enum 'Direction'
        QtMocHelpers::EnumData<enum Direction>(34, 34, QMC::EnumFlags{}).add({
            {   35, Direction::TopToBottom },
            {   36, Direction::BottomToTop },
        }),
        // enum 'YCbCrColorSpace'
        QtMocHelpers::EnumData<enum YCbCrColorSpace>(37, 37, QMC::EnumFlags{}).add({
            {   38, YCbCrColorSpace::YCbCr_Undefined },
            {   39, YCbCrColorSpace::YCbCr_BT601 },
            {   40, YCbCrColorSpace::YCbCr_BT709 },
            {   41, YCbCrColorSpace::YCbCr_xvYCC601 },
            {   42, YCbCrColorSpace::YCbCr_xvYCC709 },
            {   43, YCbCrColorSpace::YCbCr_JPEG },
            {   44, YCbCrColorSpace::YCbCr_BT2020 },
        }),
        // enum 'ColorSpace'
        QtMocHelpers::EnumData<enum ColorSpace>(45, 45, QMC::EnumFlags{}).add({
            {   46, ColorSpace::ColorSpace_Undefined },
            {   47, ColorSpace::ColorSpace_BT601 },
            {   48, ColorSpace::ColorSpace_BT709 },
            {   49, ColorSpace::ColorSpace_AdobeRgb },
            {   50, ColorSpace::ColorSpace_BT2020 },
        }),
        // enum 'ColorTransfer'
        QtMocHelpers::EnumData<enum ColorTransfer>(51, 51, QMC::EnumFlags{}).add({
            {   52, ColorTransfer::ColorTransfer_Unknown },
            {   53, ColorTransfer::ColorTransfer_BT709 },
            {   54, ColorTransfer::ColorTransfer_BT601 },
            {   55, ColorTransfer::ColorTransfer_Linear },
            {   56, ColorTransfer::ColorTransfer_Gamma22 },
            {   57, ColorTransfer::ColorTransfer_Gamma28 },
            {   58, ColorTransfer::ColorTransfer_ST2084 },
            {   59, ColorTransfer::ColorTransfer_STD_B67 },
        }),
        // enum 'ColorRange'
        QtMocHelpers::EnumData<enum ColorRange>(60, 60, QMC::EnumFlags{}).add({
            {   61, ColorRange::ColorRange_Unknown },
            {   62, ColorRange::ColorRange_Video },
            {   63, ColorRange::ColorRange_Full },
        }),
    };
    return QtMocHelpers::metaObjectData<QVideoFrameFormat, qt_meta_tag_ZN17QVideoFrameFormatE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QVideoFrameFormat::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QVideoFrameFormatE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QVideoFrameFormatE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17QVideoFrameFormatE_t>.metaTypes,
    nullptr
} };

QT_WARNING_POP
