/****************************************************************************
** Generated QML type registration code
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <QtQml/qqml.h>
#include <QtQml/qqmlmoduleregistration.h>

#if __has_include(<private/qquickimagecapture_p.h>)
#  include <private/qquickimagecapture_p.h>
#endif
#if __has_include(<private/qquickmediaplayer_p.h>)
#  include <private/qquickmediaplayer_p.h>
#endif
#if __has_include(<private/qquickscreencapture_p.h>)
#  include <private/qquickscreencapture_p.h>
#endif
#if __has_include(<private/qquicksoundeffect_p.h>)
#  include <private/qquicksoundeffect_p.h>
#endif
#if __has_include(<private/qquickvideooutput_p.h>)
#  include <private/qquickvideooutput_p.h>
#endif
#if __has_include(<private/qtmultimediaquicktypes_p.h>)
#  include <private/qtmultimediaquicktypes_p.h>
#endif


#if !defined(QT_STATIC)
#define Q_QMLTYPE_EXPORT Q_DECL_EXPORT
#else
#define Q_QMLTYPE_EXPORT
#endif
Q_QMLTYPE_EXPORT void qml_register_types_QtMultimedia()
{
    qmlRegisterModule("QtMultimedia", 5, 0);
    qmlRegisterModule("QtMultimedia", 5, 254);
    qmlRegisterModule("QtMultimedia", 6, 0);
    QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
    QMetaType::fromType<QMediaPlayer *>().id();
    qmlRegisterEnum<QMediaPlayer::PlaybackState>("QMediaPlayer::PlaybackState");
    qmlRegisterEnum<QMediaPlayer::MediaStatus>("QMediaPlayer::MediaStatus");
    qmlRegisterEnum<QMediaPlayer::Error>("QMediaPlayer::Error");
    qmlRegisterEnum<QMediaPlayer::Loops>("QMediaPlayer::Loops");
    qmlRegisterEnum<QMediaPlayer::PitchCompensationAvailability>("QMediaPlayer::PitchCompensationAvailability");
    qmlRegisterTypesAndRevisions<QQuickImageCapture>("QtMultimedia", 6);
    qmlRegisterAnonymousTypesAndRevisions<QImageCapture>("QtMultimedia", 6);
    qmlRegisterTypesAndRevisions<QQuickMediaPlayer>("QtMultimedia", 6);
    qmlRegisterTypesAndRevisions<QQuickScreenCapture>("QtMultimedia", 6);
    qmlRegisterTypesAndRevisions<QQuickSoundEffect>("QtMultimedia", 6);
    qmlRegisterTypesAndRevisions<QQuickVideoOutput>("QtMultimedia", 6);
    qmlRegisterAnonymousTypesAndRevisions<QQuickItem>("QtMultimedia", 6);
    qmlRegisterEnum<QQuickVideoOutput::FillMode>("QQuickVideoOutput::FillMode");
    qmlRegisterEnum<QQuickVideoOutput::EndOfStreamPolicy>("QQuickVideoOutput::EndOfStreamPolicy");
    qmlRegisterTypesAndRevisions<QQuickVideoSink>("QtMultimedia", 6);
    QMetaType::fromType<QSoundEffect *>().id();
    qmlRegisterEnum<QSoundEffect::Loop>("QSoundEffect::Loop");
    qmlRegisterEnum<QSoundEffect::Status>("QSoundEffect::Status");
    QMetaType::fromType<QVideoSink *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QAudioDeviceForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QAudioDeviceForeign *>().id();
    qmlRegisterEnum<QAudioDevice::Mode>("QAudioDevice::Mode");
    QMetaType::fromType<QtMultimediaPrivate::QAudioDeviceDerived>().id();
    qmlRegisterNamespaceAndRevisions(&QtMultimediaPrivate::QAudioDeviceDerived::staticMetaObject, "QtMultimedia", 6, nullptr, &QtMultimediaPrivate::QAudioDeviceNamespaceForeign::staticMetaObject, nullptr);
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QAudioInputForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QAudioInputForeign *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QAudioOutputForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QAudioOutputForeign *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QCameraDeviceForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QCameraDeviceForeign *>().id();
    qmlRegisterEnum<QCameraDevice::Position>("QCameraDevice::Position");
    QMetaType::fromType<QtMultimediaPrivate::QCameraDeviceDerived>().id();
    qmlRegisterNamespaceAndRevisions(&QtMultimediaPrivate::QCameraDeviceDerived::staticMetaObject, "QtMultimedia", 6, nullptr, &QtMultimediaPrivate::QCameraDeviceNamespaceForeign::staticMetaObject, nullptr);
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QCameraForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QCameraForeign *>().id();
    qmlRegisterEnum<QCamera::Error>("QCamera::Error");
    qmlRegisterEnum<QCamera::FocusMode>("QCamera::FocusMode");
    qmlRegisterEnum<QCamera::FlashMode>("QCamera::FlashMode");
    qmlRegisterEnum<QCamera::TorchMode>("QCamera::TorchMode");
    qmlRegisterEnum<QCamera::ExposureMode>("QCamera::ExposureMode");
    qmlRegisterEnum<QCamera::WhiteBalanceMode>("QCamera::WhiteBalanceMode");
    qmlRegisterEnum<QCamera::Features>("QCamera::Features");
    qmlRegisterEnum<QCamera::Feature>("QCamera::Feature");
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QCameraFormatForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QCameraFormatForeign *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QCapturableWindowForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QCapturableWindowForeign *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QImageCaptureForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QImageCaptureForeign *>().id();
    qmlRegisterEnum<QImageCapture::Error>("QImageCapture::Error");
    qmlRegisterEnum<QImageCapture::Quality>("QImageCapture::Quality");
    qmlRegisterEnum<QImageCapture::FileFormat>("QImageCapture::FileFormat");
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QMediaCaptureSessionForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QMediaCaptureSessionForeign *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QMediaDevicesForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QMediaDevicesForeign *>().id();
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QMediaFormatForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QMediaFormatForeign *>().id();
    qmlRegisterEnum<QMediaFormat::FileFormat>("QMediaFormat::FileFormat");
    qmlRegisterEnum<QMediaFormat::AudioCodec>("QMediaFormat::AudioCodec");
    qmlRegisterEnum<QMediaFormat::VideoCodec>("QMediaFormat::VideoCodec");
    qmlRegisterEnum<QMediaFormat::ConversionMode>("QMediaFormat::ConversionMode");
    QMetaType::fromType<QtMultimediaPrivate::QMediaFormatDerived>().id();
    qmlRegisterNamespaceAndRevisions(&QtMultimediaPrivate::QMediaFormatDerived::staticMetaObject, "QtMultimedia", 6, nullptr, &QtMultimediaPrivate::QMediaFormatNamespaceForeign::staticMetaObject, nullptr);
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QMediaMetaDataForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QMediaMetaDataForeign *>().id();
    qmlRegisterEnum<QMediaMetaData::Key>("QMediaMetaData::Key");
    QMetaType::fromType<QtMultimediaPrivate::QMediaMetaDataDerived>().id();
    qmlRegisterNamespaceAndRevisions(&QtMultimediaPrivate::QMediaMetaDataDerived::staticMetaObject, "QtMultimedia", 6, nullptr, &QtMultimediaPrivate::QMediaMetaDataNamespaceForeign::staticMetaObject, nullptr);
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QMediaRecorderForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QMediaRecorderForeign *>().id();
    qmlRegisterEnum<QMediaRecorder::Quality>("QMediaRecorder::Quality");
    qmlRegisterEnum<QMediaRecorder::EncodingMode>("QMediaRecorder::EncodingMode");
    qmlRegisterEnum<QMediaRecorder::RecorderState>("QMediaRecorder::RecorderState");
    qmlRegisterEnum<QMediaRecorder::Error>("QMediaRecorder::Error");
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QPlaybackOptionsDerived>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QPlaybackOptionsDerived *>().id();
    qmlRegisterEnum<QPlaybackOptions::PlaybackIntent>("QPlaybackOptions::PlaybackIntent");
    QMetaType::fromType<QPlaybackOptions>().id();
    qmlRegisterNamespaceAndRevisions(&QPlaybackOptions::staticMetaObject, "QtMultimedia", 6, nullptr, &QtMultimediaPrivate::QPlaybackOptionsNamespaceForeign::staticMetaObject, nullptr);
    qmlRegisterEnum<QPlaybackOptions::PlaybackIntent>("QPlaybackOptions::PlaybackIntent");
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QScreenCaptureForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QScreenCaptureForeign *>().id();
    qmlRegisterEnum<QScreenCapture::Error>("QScreenCapture::Error");
    qmlRegisterTypesAndRevisions<QtMultimediaPrivate::QWindowCaptureForeign>("QtMultimedia", 6);
    QMetaType::fromType<QtMultimediaPrivate::QWindowCaptureForeign *>().id();
    qmlRegisterEnum<QWindowCapture::Error>("QWindowCapture::Error");
    QT_WARNING_POP
    qmlRegisterModule("QtMultimedia", 6, 12);
}

static const QQmlModuleRegistration qtMultimediaRegistration("QtMultimedia", qml_register_types_QtMultimedia);
