/****************************************************************************
** Meta object code from reading C++ file 'qffmpegrecordingengine_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../../qtmultimedia/src/plugins/multimedia/ffmpeg/recordingengine/qffmpegrecordingengine_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qffmpegrecordingengine_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t {};
} // unnamed namespace

template <> constexpr inline auto QFFmpeg::RecordingEngine::qt_create_metaobjectdata<qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QFFmpeg::RecordingEngine",
        "durationChanged",
        "",
        "duration",
        "sessionError",
        "QMediaRecorder::Error",
        "code",
        "description",
        "streamInitializationError",
        "finalizationDone",
        "autoStopped",
        "newTimeStamp",
        "time"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'durationChanged'
        QtMocHelpers::SignalData<void(qint64)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 3 },
        }}),
        // Signal 'sessionError'
        QtMocHelpers::SignalData<void(QMediaRecorder::Error, const QString &)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 }, { QMetaType::QString, 7 },
        }}),
        // Signal 'streamInitializationError'
        QtMocHelpers::SignalData<void(QMediaRecorder::Error, const QString &)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 }, { QMetaType::QString, 7 },
        }}),
        // Signal 'finalizationDone'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'autoStopped'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'newTimeStamp'
        QtMocHelpers::SlotData<void(qint64)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 12 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<RecordingEngine, qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QFFmpeg::RecordingEngine::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t>.metaTypes,
    nullptr
} };

void QFFmpeg::RecordingEngine::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<RecordingEngine *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->durationChanged((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 1: _t->sessionError((*reinterpret_cast<std::add_pointer_t<QMediaRecorder::Error>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 2: _t->streamInitializationError((*reinterpret_cast<std::add_pointer_t<QMediaRecorder::Error>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 3: _t->finalizationDone(); break;
        case 4: _t->autoStopped(); break;
        case 5: _t->newTimeStamp((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (RecordingEngine::*)(qint64 )>(_a, &RecordingEngine::durationChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (RecordingEngine::*)(QMediaRecorder::Error , const QString & )>(_a, &RecordingEngine::sessionError, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (RecordingEngine::*)(QMediaRecorder::Error , const QString & )>(_a, &RecordingEngine::streamInitializationError, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (RecordingEngine::*)()>(_a, &RecordingEngine::finalizationDone, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (RecordingEngine::*)()>(_a, &RecordingEngine::autoStopped, 4))
            return;
    }
}

const QMetaObject *QFFmpeg::RecordingEngine::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QFFmpeg::RecordingEngine::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7QFFmpeg15RecordingEngineE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QFFmpeg::RecordingEngine::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    return _id;
}

// SIGNAL 0
void QFFmpeg::RecordingEngine::durationChanged(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QFFmpeg::RecordingEngine::sessionError(QMediaRecorder::Error _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void QFFmpeg::RecordingEngine::streamInitializationError(QMediaRecorder::Error _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void QFFmpeg::RecordingEngine::finalizationDone()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QFFmpeg::RecordingEngine::autoStopped()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}
QT_WARNING_POP
