/****************************************************************************
** Meta object code from reading C++ file 'qgstreamervideosink_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../../qtmultimedia/src/plugins/multimedia/gstreamer/common/qgstreamervideosink_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qgstreamervideosink_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t {};
} // unnamed namespace

template <> constexpr inline auto QGstreamerPluggableVideoSink::qt_create_metaobjectdata<qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QGstreamerPluggableVideoSink"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QGstreamerPluggableVideoSink, qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QGstreamerPluggableVideoSink::staticMetaObject = { {
    QMetaObject::SuperData::link<QPlatformVideoSink::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t>.metaTypes,
    nullptr
} };

void QGstreamerPluggableVideoSink::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QGstreamerPluggableVideoSink *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *QGstreamerPluggableVideoSink::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QGstreamerPluggableVideoSink::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN28QGstreamerPluggableVideoSinkE_t>.strings))
        return static_cast<void*>(this);
    return QPlatformVideoSink::qt_metacast(_clname);
}

int QGstreamerPluggableVideoSink::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QPlatformVideoSink::qt_metacall(_c, _id, _a);
    return _id;
}
namespace {
struct qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t {};
} // unnamed namespace

template <> constexpr inline auto QGstreamerRelayVideoSink::qt_create_metaobjectdata<qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QGstreamerRelayVideoSink",
        "aboutToBeDestroyed",
        "",
        "videoFrameChanged",
        "QVideoFrame",
        "frame",
        "subtitleTextChanged",
        "subtitleText",
        "nativeSizeChanged",
        "QSize",
        "size"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'aboutToBeDestroyed'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoFrameChanged'
        QtMocHelpers::SignalData<void(const QVideoFrame &)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'subtitleTextChanged'
        QtMocHelpers::SignalData<void(const QString &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 7 },
        }}),
        // Signal 'nativeSizeChanged'
        QtMocHelpers::SignalData<void(QSize)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 10 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QGstreamerRelayVideoSink, qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QGstreamerRelayVideoSink::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t>.metaTypes,
    nullptr
} };

void QGstreamerRelayVideoSink::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QGstreamerRelayVideoSink *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->aboutToBeDestroyed(); break;
        case 1: _t->videoFrameChanged((*reinterpret_cast<std::add_pointer_t<QVideoFrame>>(_a[1]))); break;
        case 2: _t->subtitleTextChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 3: _t->nativeSizeChanged((*reinterpret_cast<std::add_pointer_t<QSize>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 1:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QVideoFrame >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QGstreamerRelayVideoSink::*)()>(_a, &QGstreamerRelayVideoSink::aboutToBeDestroyed, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QGstreamerRelayVideoSink::*)(const QVideoFrame & )>(_a, &QGstreamerRelayVideoSink::videoFrameChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QGstreamerRelayVideoSink::*)(const QString & )>(_a, &QGstreamerRelayVideoSink::subtitleTextChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QGstreamerRelayVideoSink::*)(QSize )>(_a, &QGstreamerRelayVideoSink::nativeSizeChanged, 3))
            return;
    }
}

const QMetaObject *QGstreamerRelayVideoSink::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QGstreamerRelayVideoSink::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN24QGstreamerRelayVideoSinkE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QGstreamerRelayVideoSink::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void QGstreamerRelayVideoSink::aboutToBeDestroyed()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QGstreamerRelayVideoSink::videoFrameChanged(const QVideoFrame & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QGstreamerRelayVideoSink::subtitleTextChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QGstreamerRelayVideoSink::nativeSizeChanged(QSize _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}
QT_WARNING_POP
