/****************************************************************************
** Meta object code from reading C++ file 'qspatialsound.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/spatialaudio/qspatialsound.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qspatialsound.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QSpatialSoundE_t {};
} // unnamed namespace

template <> constexpr inline auto QSpatialSound::qt_create_metaobjectdata<qt_meta_tag_ZN13QSpatialSoundE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QSpatialSound",
        "sourceChanged",
        "",
        "loopsChanged",
        "autoPlayChanged",
        "positionChanged",
        "rotationChanged",
        "volumeChanged",
        "distanceModelChanged",
        "sizeChanged",
        "distanceCutoffChanged",
        "manualAttenuationChanged",
        "occlusionIntensityChanged",
        "directivityChanged",
        "directivityOrderChanged",
        "nearFieldGainChanged",
        "play",
        "pause",
        "stop",
        "source",
        "QUrl",
        "position",
        "QVector3D",
        "rotation",
        "QQuaternion",
        "volume",
        "distanceModel",
        "DistanceModel",
        "size",
        "distanceCutoff",
        "manualAttenuation",
        "occlusionIntensity",
        "directivity",
        "directivityOrder",
        "nearFieldGain",
        "loops",
        "autoPlay",
        "Loops",
        "Infinite",
        "Once",
        "Logarithmic",
        "Linear",
        "ManualAttenuation"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sourceChanged'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'loopsChanged'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'autoPlayChanged'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'rotationChanged'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'volumeChanged'
        QtMocHelpers::SignalData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'distanceModelChanged'
        QtMocHelpers::SignalData<void()>(8, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'sizeChanged'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'distanceCutoffChanged'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'manualAttenuationChanged'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'occlusionIntensityChanged'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'directivityChanged'
        QtMocHelpers::SignalData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'directivityOrderChanged'
        QtMocHelpers::SignalData<void()>(14, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'nearFieldGainChanged'
        QtMocHelpers::SignalData<void()>(15, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'pause'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'source'
        QtMocHelpers::PropertyData<QUrl>(19, 0x80000000 | 20, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'position'
        QtMocHelpers::PropertyData<QVector3D>(21, 0x80000000 | 22, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'rotation'
        QtMocHelpers::PropertyData<QQuaternion>(23, 0x80000000 | 24, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 4),
        // property 'volume'
        QtMocHelpers::PropertyData<float>(25, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'distanceModel'
        QtMocHelpers::PropertyData<enum DistanceModel>(26, 0x80000000 | 27, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 6),
        // property 'size'
        QtMocHelpers::PropertyData<float>(28, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 7),
        // property 'distanceCutoff'
        QtMocHelpers::PropertyData<float>(29, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 8),
        // property 'manualAttenuation'
        QtMocHelpers::PropertyData<float>(30, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 9),
        // property 'occlusionIntensity'
        QtMocHelpers::PropertyData<float>(31, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 10),
        // property 'directivity'
        QtMocHelpers::PropertyData<float>(32, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 11),
        // property 'directivityOrder'
        QtMocHelpers::PropertyData<float>(33, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 12),
        // property 'nearFieldGain'
        QtMocHelpers::PropertyData<float>(34, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 13),
        // property 'loops'
        QtMocHelpers::PropertyData<int>(35, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 1),
        // property 'autoPlay'
        QtMocHelpers::PropertyData<bool>(36, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Loops'
        QtMocHelpers::EnumData<enum Loops>(37, 37, QMC::EnumFlags{}).add({
            {   38, Loops::Infinite },
            {   39, Loops::Once },
        }),
        // enum 'DistanceModel'
        QtMocHelpers::EnumData<enum DistanceModel>(27, 27, QMC::EnumIsScoped).add({
            {   40, DistanceModel::Logarithmic },
            {   41, DistanceModel::Linear },
            {   42, DistanceModel::ManualAttenuation },
        }),
    };
    return QtMocHelpers::metaObjectData<QSpatialSound, qt_meta_tag_ZN13QSpatialSoundE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QSpatialSound::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QSpatialSoundE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QSpatialSoundE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QSpatialSoundE_t>.metaTypes,
    nullptr
} };

void QSpatialSound::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QSpatialSound *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sourceChanged(); break;
        case 1: _t->loopsChanged(); break;
        case 2: _t->autoPlayChanged(); break;
        case 3: _t->positionChanged(); break;
        case 4: _t->rotationChanged(); break;
        case 5: _t->volumeChanged(); break;
        case 6: _t->distanceModelChanged(); break;
        case 7: _t->sizeChanged(); break;
        case 8: _t->distanceCutoffChanged(); break;
        case 9: _t->manualAttenuationChanged(); break;
        case 10: _t->occlusionIntensityChanged(); break;
        case 11: _t->directivityChanged(); break;
        case 12: _t->directivityOrderChanged(); break;
        case 13: _t->nearFieldGainChanged(); break;
        case 14: _t->play(); break;
        case 15: _t->pause(); break;
        case 16: _t->stop(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::sourceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::loopsChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::autoPlayChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::positionChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::rotationChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::volumeChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::distanceModelChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::sizeChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::distanceCutoffChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::manualAttenuationChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::occlusionIntensityChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::directivityChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::directivityOrderChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSpatialSound::*)()>(_a, &QSpatialSound::nearFieldGainChanged, 13))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QUrl*>(_v) = _t->source(); break;
        case 1: *reinterpret_cast<QVector3D*>(_v) = _t->position(); break;
        case 2: *reinterpret_cast<QQuaternion*>(_v) = _t->rotation(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->volume(); break;
        case 4: *reinterpret_cast<enum DistanceModel*>(_v) = _t->distanceModel(); break;
        case 5: *reinterpret_cast<float*>(_v) = _t->size(); break;
        case 6: *reinterpret_cast<float*>(_v) = _t->distanceCutoff(); break;
        case 7: *reinterpret_cast<float*>(_v) = _t->manualAttenuation(); break;
        case 8: *reinterpret_cast<float*>(_v) = _t->occlusionIntensity(); break;
        case 9: *reinterpret_cast<float*>(_v) = _t->directivity(); break;
        case 10: *reinterpret_cast<float*>(_v) = _t->directivityOrder(); break;
        case 11: *reinterpret_cast<float*>(_v) = _t->nearFieldGain(); break;
        case 12: *reinterpret_cast<int*>(_v) = _t->loops(); break;
        case 13: *reinterpret_cast<bool*>(_v) = _t->autoPlay(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setSource(*reinterpret_cast<QUrl*>(_v)); break;
        case 1: _t->setPosition(*reinterpret_cast<QVector3D*>(_v)); break;
        case 2: _t->setRotation(*reinterpret_cast<QQuaternion*>(_v)); break;
        case 3: _t->setVolume(*reinterpret_cast<float*>(_v)); break;
        case 4: _t->setDistanceModel(*reinterpret_cast<enum DistanceModel*>(_v)); break;
        case 5: _t->setSize(*reinterpret_cast<float*>(_v)); break;
        case 6: _t->setDistanceCutoff(*reinterpret_cast<float*>(_v)); break;
        case 7: _t->setManualAttenuation(*reinterpret_cast<float*>(_v)); break;
        case 8: _t->setOcclusionIntensity(*reinterpret_cast<float*>(_v)); break;
        case 9: _t->setDirectivity(*reinterpret_cast<float*>(_v)); break;
        case 10: _t->setDirectivityOrder(*reinterpret_cast<float*>(_v)); break;
        case 11: _t->setNearFieldGain(*reinterpret_cast<float*>(_v)); break;
        case 12: _t->setLoops(*reinterpret_cast<int*>(_v)); break;
        case 13: _t->setAutoPlay(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QSpatialSound::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QSpatialSound::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QSpatialSoundE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QSpatialSound::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 17)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 17;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 17)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 17;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    }
    return _id;
}

// SIGNAL 0
void QSpatialSound::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QSpatialSound::loopsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QSpatialSound::autoPlayChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QSpatialSound::positionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QSpatialSound::rotationChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QSpatialSound::volumeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QSpatialSound::distanceModelChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QSpatialSound::sizeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void QSpatialSound::distanceCutoffChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void QSpatialSound::manualAttenuationChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void QSpatialSound::occlusionIntensityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 10, nullptr);
}

// SIGNAL 11
void QSpatialSound::directivityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void QSpatialSound::directivityOrderChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 12, nullptr);
}

// SIGNAL 13
void QSpatialSound::nearFieldGainChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}
QT_WARNING_POP
