/****************************************************************************
** Meta object code from reading C++ file 'qsoundeffect.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/audio/qsoundeffect.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qsoundeffect.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN12QSoundEffectE_t {};
} // unnamed namespace

template <> constexpr inline auto QSoundEffect::qt_create_metaobjectdata<qt_meta_tag_ZN12QSoundEffectE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QSoundEffect",
        "DefaultMethod",
        "play()",
        "sourceChanged",
        "",
        "loopCountChanged",
        "loopsRemainingChanged",
        "volumeChanged",
        "mutedChanged",
        "loadedChanged",
        "playingChanged",
        "statusChanged",
        "audioDeviceChanged",
        "play",
        "stop",
        "source",
        "QUrl",
        "loops",
        "loopsRemaining",
        "volume",
        "muted",
        "playing",
        "status",
        "Status",
        "audioDevice",
        "QAudioDevice",
        "Loop",
        "Infinite",
        "Null",
        "Loading",
        "Ready",
        "Error"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sourceChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'loopCountChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'loopsRemainingChanged'
        QtMocHelpers::SignalData<void()>(6, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'volumeChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'mutedChanged'
        QtMocHelpers::SignalData<void()>(8, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'loadedChanged'
        QtMocHelpers::SignalData<void()>(9, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playingChanged'
        QtMocHelpers::SignalData<void()>(10, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'statusChanged'
        QtMocHelpers::SignalData<void()>(11, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioDeviceChanged'
        QtMocHelpers::SignalData<void()>(12, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(13, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(14, 4, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'source'
        QtMocHelpers::PropertyData<QUrl>(15, 0x80000000 | 16, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'loops'
        QtMocHelpers::PropertyData<int>(17, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable, 1),
        // property 'loopsRemaining'
        QtMocHelpers::PropertyData<int>(18, QMetaType::Int, QMC::DefaultPropertyFlags, 2),
        // property 'volume'
        QtMocHelpers::PropertyData<float>(19, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3),
        // property 'muted'
        QtMocHelpers::PropertyData<bool>(20, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'playing'
        QtMocHelpers::PropertyData<bool>(21, QMetaType::Bool, QMC::DefaultPropertyFlags, 6),
        // property 'status'
        QtMocHelpers::PropertyData<enum Status>(22, 0x80000000 | 23, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 7),
        // property 'audioDevice'
        QtMocHelpers::PropertyData<QAudioDevice>(24, 0x80000000 | 25, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 8),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Loop'
        QtMocHelpers::EnumData<enum Loop>(26, 26, QMC::EnumFlags{}).add({
            {   27, Loop::Infinite },
        }),
        // enum 'Status'
        QtMocHelpers::EnumData<enum Status>(23, 23, QMC::EnumFlags{}).add({
            {   28, Status::Null },
            {   29, Status::Loading },
            {   30, Status::Ready },
            {   31, Status::Error },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QSoundEffect, qt_meta_tag_ZN12QSoundEffectE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QSoundEffect::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QSoundEffectE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QSoundEffectE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN12QSoundEffectE_t>.metaTypes,
    nullptr
} };

void QSoundEffect::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QSoundEffect *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sourceChanged(); break;
        case 1: _t->loopCountChanged(); break;
        case 2: _t->loopsRemainingChanged(); break;
        case 3: _t->volumeChanged(); break;
        case 4: _t->mutedChanged(); break;
        case 5: _t->loadedChanged(); break;
        case 6: _t->playingChanged(); break;
        case 7: _t->statusChanged(); break;
        case 8: _t->audioDeviceChanged(); break;
        case 9: _t->play(); break;
        case 10: _t->stop(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::sourceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::loopCountChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::loopsRemainingChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::volumeChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::mutedChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::loadedChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::playingChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::statusChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QSoundEffect::*)()>(_a, &QSoundEffect::audioDeviceChanged, 8))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QUrl*>(_v) = _t->source(); break;
        case 1: *reinterpret_cast<int*>(_v) = _t->loopCount(); break;
        case 2: *reinterpret_cast<int*>(_v) = _t->loopsRemaining(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->volume(); break;
        case 4: *reinterpret_cast<bool*>(_v) = _t->isMuted(); break;
        case 5: *reinterpret_cast<bool*>(_v) = _t->isPlaying(); break;
        case 6: *reinterpret_cast<enum Status*>(_v) = _t->status(); break;
        case 7: *reinterpret_cast<QAudioDevice*>(_v) = _t->audioDevice(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setSource(*reinterpret_cast<QUrl*>(_v)); break;
        case 1: _t->setLoopCount(*reinterpret_cast<int*>(_v)); break;
        case 3: _t->setVolume(*reinterpret_cast<float*>(_v)); break;
        case 4: _t->setMuted(*reinterpret_cast<bool*>(_v)); break;
        case 7: _t->setAudioDevice(*reinterpret_cast<QAudioDevice*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QSoundEffect::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QSoundEffect::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QSoundEffectE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QSoundEffect::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 11;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    return _id;
}

// SIGNAL 0
void QSoundEffect::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QSoundEffect::loopCountChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QSoundEffect::loopsRemainingChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QSoundEffect::volumeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QSoundEffect::mutedChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QSoundEffect::loadedChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QSoundEffect::playingChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QSoundEffect::statusChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void QSoundEffect::audioDeviceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}
QT_WARNING_POP
