/****************************************************************************
** Meta object code from reading C++ file 'qmediaplayer.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/playback/qmediaplayer.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qmediaplayer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN12QMediaPlayerE_t {};
} // unnamed namespace

template <> constexpr inline auto QMediaPlayer::qt_create_metaobjectdata<qt_meta_tag_ZN12QMediaPlayerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QMediaPlayer",
        "sourceChanged",
        "",
        "QUrl",
        "media",
        "playbackStateChanged",
        "QMediaPlayer::PlaybackState",
        "newState",
        "mediaStatusChanged",
        "QMediaPlayer::MediaStatus",
        "status",
        "durationChanged",
        "duration",
        "positionChanged",
        "position",
        "hasAudioChanged",
        "available",
        "hasVideoChanged",
        "videoAvailable",
        "bufferProgressChanged",
        "progress",
        "seekableChanged",
        "seekable",
        "playingChanged",
        "playing",
        "playbackRateChanged",
        "rate",
        "loopsChanged",
        "metaDataChanged",
        "videoOutputChanged",
        "audioOutputChanged",
        "audioBufferOutputChanged",
        "tracksChanged",
        "activeTracksChanged",
        "errorChanged",
        "errorOccurred",
        "QMediaPlayer::Error",
        "error",
        "errorString",
        "pitchCompensationChanged",
        "playbackOptionsChanged",
        "play",
        "pause",
        "stop",
        "setPosition",
        "setPlaybackRate",
        "setSource",
        "source",
        "setSourceDevice",
        "QIODevice*",
        "device",
        "sourceUrl",
        "setPitchCompensation",
        "setPlaybackOptions",
        "QPlaybackOptions",
        "options",
        "resetPlaybackOptions",
        "bufferProgress",
        "hasAudio",
        "hasVideo",
        "playbackRate",
        "loops",
        "playbackState",
        "PlaybackState",
        "mediaStatus",
        "MediaStatus",
        "metaData",
        "QMediaMetaData",
        "Error",
        "videoOutput",
        "audioOutput",
        "QAudioOutput*",
        "audioBufferOutput",
        "QAudioBufferOutput*",
        "audioTracks",
        "QList<QMediaMetaData>",
        "videoTracks",
        "subtitleTracks",
        "activeAudioTrack",
        "activeVideoTrack",
        "activeSubtitleTrack",
        "pitchCompensationAvailability",
        "PitchCompensationAvailability",
        "pitchCompensation",
        "playbackOptions",
        "StoppedState",
        "PlayingState",
        "PausedState",
        "NoMedia",
        "LoadingMedia",
        "LoadedMedia",
        "StalledMedia",
        "BufferingMedia",
        "BufferedMedia",
        "EndOfMedia",
        "InvalidMedia",
        "NoError",
        "ResourceError",
        "FormatError",
        "NetworkError",
        "AccessDeniedError",
        "Loops",
        "Infinite",
        "Once",
        "AlwaysOn",
        "Available",
        "Unavailable"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sourceChanged'
        QtMocHelpers::SignalData<void(const QUrl &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'playbackStateChanged'
        QtMocHelpers::SignalData<void(QMediaPlayer::PlaybackState)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 7 },
        }}),
        // Signal 'mediaStatusChanged'
        QtMocHelpers::SignalData<void(QMediaPlayer::MediaStatus)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 10 },
        }}),
        // Signal 'durationChanged'
        QtMocHelpers::SignalData<void(qint64)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 12 },
        }}),
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void(qint64)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 14 },
        }}),
        // Signal 'hasAudioChanged'
        QtMocHelpers::SignalData<void(bool)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 16 },
        }}),
        // Signal 'hasVideoChanged'
        QtMocHelpers::SignalData<void(bool)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 18 },
        }}),
        // Signal 'bufferProgressChanged'
        QtMocHelpers::SignalData<void(float)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 20 },
        }}),
        // Signal 'seekableChanged'
        QtMocHelpers::SignalData<void(bool)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 22 },
        }}),
        // Signal 'playingChanged'
        QtMocHelpers::SignalData<void(bool)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 24 },
        }}),
        // Signal 'playbackRateChanged'
        QtMocHelpers::SignalData<void(qreal)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QReal, 26 },
        }}),
        // Signal 'loopsChanged'
        QtMocHelpers::SignalData<void()>(27, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'metaDataChanged'
        QtMocHelpers::SignalData<void()>(28, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoOutputChanged'
        QtMocHelpers::SignalData<void()>(29, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioOutputChanged'
        QtMocHelpers::SignalData<void()>(30, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioBufferOutputChanged'
        QtMocHelpers::RevisionedSignalData<void()>(31, 2, QMC::AccessPublic, 0x608, QMetaType::Void),
        // Signal 'tracksChanged'
        QtMocHelpers::SignalData<void()>(32, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'activeTracksChanged'
        QtMocHelpers::SignalData<void()>(33, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'errorChanged'
        QtMocHelpers::SignalData<void()>(34, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'errorOccurred'
        QtMocHelpers::SignalData<void(QMediaPlayer::Error, const QString &)>(35, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 36, 37 }, { QMetaType::QString, 38 },
        }}),
        // Signal 'pitchCompensationChanged'
        QtMocHelpers::RevisionedSignalData<void(bool)>(39, 2, QMC::AccessPublic, 0x60a, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'playbackOptionsChanged'
        QtMocHelpers::RevisionedSignalData<void()>(40, 2, QMC::AccessPublic, 0x60a, QMetaType::Void),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(41, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'pause'
        QtMocHelpers::SlotData<void()>(42, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'setPosition'
        QtMocHelpers::SlotData<void(qint64)>(44, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 14 },
        }}),
        // Slot 'setPlaybackRate'
        QtMocHelpers::SlotData<void(qreal)>(45, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QReal, 26 },
        }}),
        // Slot 'setSource'
        QtMocHelpers::SlotData<void(const QUrl &)>(46, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 47 },
        }}),
        // Slot 'setSourceDevice'
        QtMocHelpers::SlotData<void(QIODevice *, const QUrl &)>(48, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 49, 50 }, { 0x80000000 | 3, 51 },
        }}),
        // Slot 'setSourceDevice'
        QtMocHelpers::SlotData<void(QIODevice *)>(48, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { 0x80000000 | 49, 50 },
        }}),
        // Slot 'setPitchCompensation'
        QtMocHelpers::SlotData<void(bool) const>(52, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setPlaybackOptions'
        QtMocHelpers::SlotData<void(const QPlaybackOptions &)>(53, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 54, 55 },
        }}),
        // Slot 'resetPlaybackOptions'
        QtMocHelpers::SlotData<void()>(56, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'source'
        QtMocHelpers::PropertyData<QUrl>(47, 0x80000000 | 3, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Virtual, 0),
        // property 'duration'
        QtMocHelpers::PropertyData<qint64>(12, QMetaType::LongLong, QMC::DefaultPropertyFlags, 3),
        // property 'position'
        QtMocHelpers::PropertyData<qint64>(14, QMetaType::LongLong, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'bufferProgress'
        QtMocHelpers::PropertyData<float>(57, QMetaType::Float, QMC::DefaultPropertyFlags, 7),
        // property 'hasAudio'
        QtMocHelpers::PropertyData<bool>(58, QMetaType::Bool, QMC::DefaultPropertyFlags, 5),
        // property 'hasVideo'
        QtMocHelpers::PropertyData<bool>(59, QMetaType::Bool, QMC::DefaultPropertyFlags, 6),
        // property 'seekable'
        QtMocHelpers::PropertyData<bool>(22, QMetaType::Bool, QMC::DefaultPropertyFlags, 8),
        // property 'playing'
        QtMocHelpers::PropertyData<bool>(24, QMetaType::Bool, QMC::DefaultPropertyFlags, 9),
        // property 'playbackRate'
        QtMocHelpers::PropertyData<qreal>(60, QMetaType::QReal, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 10),
        // property 'loops'
        QtMocHelpers::PropertyData<int>(61, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 11),
        // property 'playbackState'
        QtMocHelpers::PropertyData<enum PlaybackState>(62, 0x80000000 | 63, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 1),
        // property 'mediaStatus'
        QtMocHelpers::PropertyData<enum MediaStatus>(64, 0x80000000 | 65, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 2),
        // property 'metaData'
        QtMocHelpers::PropertyData<QMediaMetaData>(66, 0x80000000 | 67, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 12),
        // property 'error'
        QtMocHelpers::PropertyData<enum Error>(37, 0x80000000 | 68, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 18),
        // property 'errorString'
        QtMocHelpers::PropertyData<QString>(38, QMetaType::QString, QMC::DefaultPropertyFlags, 18),
        // property 'videoOutput'
        QtMocHelpers::PropertyData<QObject*>(69, QMetaType::QObjectStar, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 13),
        // property 'audioOutput'
        QtMocHelpers::PropertyData<QAudioOutput*>(70, 0x80000000 | 71, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 14),
        // property 'audioBufferOutput'
        QtMocHelpers::PropertyData<QAudioBufferOutput*>(72, 0x80000000 | 73, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 15),
        // property 'audioTracks'
        QtMocHelpers::PropertyData<QList<QMediaMetaData>>(74, 0x80000000 | 75, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 16),
        // property 'videoTracks'
        QtMocHelpers::PropertyData<QList<QMediaMetaData>>(76, 0x80000000 | 75, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 16),
        // property 'subtitleTracks'
        QtMocHelpers::PropertyData<QList<QMediaMetaData>>(77, 0x80000000 | 75, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 16),
        // property 'activeAudioTrack'
        QtMocHelpers::PropertyData<int>(78, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 17),
        // property 'activeVideoTrack'
        QtMocHelpers::PropertyData<int>(79, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 17),
        // property 'activeSubtitleTrack'
        QtMocHelpers::PropertyData<int>(80, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 17),
        // property 'pitchCompensationAvailability'
        QtMocHelpers::PropertyData<enum PitchCompensationAvailability>(81, 0x80000000 | 82, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'pitchCompensation'
        QtMocHelpers::PropertyData<bool>(83, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 20),
        // property 'playbackOptions'
        QtMocHelpers::PropertyData<QPlaybackOptions>(84, 0x80000000 | 54, QMC::DefaultPropertyFlags | QMC::Writable | QMC::Resettable | QMC::EnumOrFlag | QMC::StdCppSet, 21),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'PlaybackState'
        QtMocHelpers::EnumData<enum PlaybackState>(63, 63, QMC::EnumFlags{}).add({
            {   85, PlaybackState::StoppedState },
            {   86, PlaybackState::PlayingState },
            {   87, PlaybackState::PausedState },
        }),
        // enum 'MediaStatus'
        QtMocHelpers::EnumData<enum MediaStatus>(65, 65, QMC::EnumFlags{}).add({
            {   88, MediaStatus::NoMedia },
            {   89, MediaStatus::LoadingMedia },
            {   90, MediaStatus::LoadedMedia },
            {   91, MediaStatus::StalledMedia },
            {   92, MediaStatus::BufferingMedia },
            {   93, MediaStatus::BufferedMedia },
            {   94, MediaStatus::EndOfMedia },
            {   95, MediaStatus::InvalidMedia },
        }),
        // enum 'Error'
        QtMocHelpers::EnumData<enum Error>(68, 68, QMC::EnumFlags{}).add({
            {   96, Error::NoError },
            {   97, Error::ResourceError },
            {   98, Error::FormatError },
            {   99, Error::NetworkError },
            {  100, Error::AccessDeniedError },
        }),
        // enum 'Loops'
        QtMocHelpers::EnumData<enum Loops>(101, 101, QMC::EnumFlags{}).add({
            {  102, Loops::Infinite },
            {  103, Loops::Once },
        }),
        // enum 'PitchCompensationAvailability'
        QtMocHelpers::EnumData<enum PitchCompensationAvailability>(82, 82, QMC::EnumIsScoped).add({
            {  104, PitchCompensationAvailability::AlwaysOn },
            {  105, PitchCompensationAvailability::Available },
            {  106, PitchCompensationAvailability::Unavailable },
        }),
    };
    return QtMocHelpers::metaObjectData<QMediaPlayer, qt_meta_tag_ZN12QMediaPlayerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QMediaPlayer::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QMediaPlayerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QMediaPlayerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN12QMediaPlayerE_t>.metaTypes,
    nullptr
} };

void QMediaPlayer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QMediaPlayer *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sourceChanged((*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[1]))); break;
        case 1: _t->playbackStateChanged((*reinterpret_cast<std::add_pointer_t<QMediaPlayer::PlaybackState>>(_a[1]))); break;
        case 2: _t->mediaStatusChanged((*reinterpret_cast<std::add_pointer_t<QMediaPlayer::MediaStatus>>(_a[1]))); break;
        case 3: _t->durationChanged((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 4: _t->positionChanged((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 5: _t->hasAudioChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 6: _t->hasVideoChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 7: _t->bufferProgressChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 8: _t->seekableChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 9: _t->playingChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 10: _t->playbackRateChanged((*reinterpret_cast<std::add_pointer_t<qreal>>(_a[1]))); break;
        case 11: _t->loopsChanged(); break;
        case 12: _t->metaDataChanged(); break;
        case 13: _t->videoOutputChanged(); break;
        case 14: _t->audioOutputChanged(); break;
        case 15: _t->audioBufferOutputChanged(); break;
        case 16: _t->tracksChanged(); break;
        case 17: _t->activeTracksChanged(); break;
        case 18: _t->errorChanged(); break;
        case 19: _t->errorOccurred((*reinterpret_cast<std::add_pointer_t<QMediaPlayer::Error>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 20: _t->pitchCompensationChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 21: _t->playbackOptionsChanged(); break;
        case 22: _t->play(); break;
        case 23: _t->pause(); break;
        case 24: _t->stop(); break;
        case 25: _t->setPosition((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 26: _t->setPlaybackRate((*reinterpret_cast<std::add_pointer_t<qreal>>(_a[1]))); break;
        case 27: _t->setSource((*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[1]))); break;
        case 28: _t->setSourceDevice((*reinterpret_cast<std::add_pointer_t<QIODevice*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[2]))); break;
        case 29: _t->setSourceDevice((*reinterpret_cast<std::add_pointer_t<QIODevice*>>(_a[1]))); break;
        case 30: _t->setPitchCompensation((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 31: _t->setPlaybackOptions((*reinterpret_cast<std::add_pointer_t<QPlaybackOptions>>(_a[1]))); break;
        case 32: _t->resetPlaybackOptions(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(const QUrl & )>(_a, &QMediaPlayer::sourceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(QMediaPlayer::PlaybackState )>(_a, &QMediaPlayer::playbackStateChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(QMediaPlayer::MediaStatus )>(_a, &QMediaPlayer::mediaStatusChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(qint64 )>(_a, &QMediaPlayer::durationChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(qint64 )>(_a, &QMediaPlayer::positionChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(bool )>(_a, &QMediaPlayer::hasAudioChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(bool )>(_a, &QMediaPlayer::hasVideoChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(float )>(_a, &QMediaPlayer::bufferProgressChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(bool )>(_a, &QMediaPlayer::seekableChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(bool )>(_a, &QMediaPlayer::playingChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(qreal )>(_a, &QMediaPlayer::playbackRateChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::loopsChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::metaDataChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::videoOutputChanged, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::audioOutputChanged, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::audioBufferOutputChanged, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::tracksChanged, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::activeTracksChanged, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::errorChanged, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(QMediaPlayer::Error , const QString & )>(_a, &QMediaPlayer::errorOccurred, 19))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)(bool )>(_a, &QMediaPlayer::pitchCompensationChanged, 20))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaPlayer::*)()>(_a, &QMediaPlayer::playbackOptionsChanged, 21))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QUrl*>(_v) = _t->source(); break;
        case 1: *reinterpret_cast<qint64*>(_v) = _t->duration(); break;
        case 2: *reinterpret_cast<qint64*>(_v) = _t->position(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->bufferProgress(); break;
        case 4: *reinterpret_cast<bool*>(_v) = _t->hasAudio(); break;
        case 5: *reinterpret_cast<bool*>(_v) = _t->hasVideo(); break;
        case 6: *reinterpret_cast<bool*>(_v) = _t->isSeekable(); break;
        case 7: *reinterpret_cast<bool*>(_v) = _t->isPlaying(); break;
        case 8: *reinterpret_cast<qreal*>(_v) = _t->playbackRate(); break;
        case 9: *reinterpret_cast<int*>(_v) = _t->loops(); break;
        case 10: *reinterpret_cast<enum PlaybackState*>(_v) = _t->playbackState(); break;
        case 11: *reinterpret_cast<enum MediaStatus*>(_v) = _t->mediaStatus(); break;
        case 12: *reinterpret_cast<QMediaMetaData*>(_v) = _t->metaData(); break;
        case 13: *reinterpret_cast<enum Error*>(_v) = _t->error(); break;
        case 14: *reinterpret_cast<QString*>(_v) = _t->errorString(); break;
        case 15: *reinterpret_cast<QObject**>(_v) = _t->videoOutput(); break;
        case 16: *reinterpret_cast<QAudioOutput**>(_v) = _t->audioOutput(); break;
        case 17: *reinterpret_cast<QAudioBufferOutput**>(_v) = _t->audioBufferOutput(); break;
        case 18: *reinterpret_cast<QList<QMediaMetaData>*>(_v) = _t->audioTracks(); break;
        case 19: *reinterpret_cast<QList<QMediaMetaData>*>(_v) = _t->videoTracks(); break;
        case 20: *reinterpret_cast<QList<QMediaMetaData>*>(_v) = _t->subtitleTracks(); break;
        case 21: *reinterpret_cast<int*>(_v) = _t->activeAudioTrack(); break;
        case 22: *reinterpret_cast<int*>(_v) = _t->activeVideoTrack(); break;
        case 23: *reinterpret_cast<int*>(_v) = _t->activeSubtitleTrack(); break;
        case 24: *reinterpret_cast<enum PitchCompensationAvailability*>(_v) = _t->pitchCompensationAvailability(); break;
        case 25: *reinterpret_cast<bool*>(_v) = _t->pitchCompensation(); break;
        case 26: *reinterpret_cast<QPlaybackOptions*>(_v) = _t->playbackOptions(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setSource(*reinterpret_cast<QUrl*>(_v)); break;
        case 2: _t->setPosition(*reinterpret_cast<qint64*>(_v)); break;
        case 8: _t->setPlaybackRate(*reinterpret_cast<qreal*>(_v)); break;
        case 9: _t->setLoops(*reinterpret_cast<int*>(_v)); break;
        case 15: _t->setVideoOutput(*reinterpret_cast<QObject**>(_v)); break;
        case 16: _t->setAudioOutput(*reinterpret_cast<QAudioOutput**>(_v)); break;
        case 17: _t->setAudioBufferOutput(*reinterpret_cast<QAudioBufferOutput**>(_v)); break;
        case 21: _t->setActiveAudioTrack(*reinterpret_cast<int*>(_v)); break;
        case 22: _t->setActiveVideoTrack(*reinterpret_cast<int*>(_v)); break;
        case 23: _t->setActiveSubtitleTrack(*reinterpret_cast<int*>(_v)); break;
        case 25: _t->setPitchCompensation(*reinterpret_cast<bool*>(_v)); break;
        case 26: _t->setPlaybackOptions(*reinterpret_cast<QPlaybackOptions*>(_v)); break;
        default: break;
        }
    }
    if (_c == QMetaObject::ResetProperty) {
        switch (_id) {
        case 26: _t->resetPlaybackOptions(); break;
        default: break;
        }
    }
}

const QMetaObject *QMediaPlayer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QMediaPlayer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QMediaPlayerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QMediaPlayer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 33)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 33;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 33)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 33;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 27;
    }
    return _id;
}

// SIGNAL 0
void QMediaPlayer::sourceChanged(const QUrl & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QMediaPlayer::playbackStateChanged(QMediaPlayer::PlaybackState _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QMediaPlayer::mediaStatusChanged(QMediaPlayer::MediaStatus _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QMediaPlayer::durationChanged(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void QMediaPlayer::positionChanged(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void QMediaPlayer::hasAudioChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void QMediaPlayer::hasVideoChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void QMediaPlayer::bufferProgressChanged(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void QMediaPlayer::seekableChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void QMediaPlayer::playingChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void QMediaPlayer::playbackRateChanged(qreal _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void QMediaPlayer::loopsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void QMediaPlayer::metaDataChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 12, nullptr);
}

// SIGNAL 13
void QMediaPlayer::videoOutputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void QMediaPlayer::audioOutputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 14, nullptr);
}

// SIGNAL 15
void QMediaPlayer::audioBufferOutputChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 15, nullptr);
}

// SIGNAL 16
void QMediaPlayer::tracksChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 16, nullptr);
}

// SIGNAL 17
void QMediaPlayer::activeTracksChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 17, nullptr);
}

// SIGNAL 18
void QMediaPlayer::errorChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 18, nullptr);
}

// SIGNAL 19
void QMediaPlayer::errorOccurred(QMediaPlayer::Error _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 19, nullptr, _t1, _t2);
}

// SIGNAL 20
void QMediaPlayer::pitchCompensationChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 20, nullptr, _t1);
}

// SIGNAL 21
void QMediaPlayer::playbackOptionsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 21, nullptr);
}
QT_WARNING_POP
