/****************************************************************************
** Meta object code from reading C++ file 'qquick3daudioroom_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/spatialaudioquick3d/qquick3daudioroom_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qquick3daudioroom_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN17QQuick3DAudioRoomE_t {};
} // unnamed namespace

template <> constexpr inline auto QQuick3DAudioRoom::qt_create_metaobjectdata<qt_meta_tag_ZN17QQuick3DAudioRoomE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QQuick3DAudioRoom",
        "QML.Element",
        "AudioRoom",
        "positionChanged",
        "",
        "dimensionsChanged",
        "rotationChanged",
        "wallsChanged",
        "reflectionGainChanged",
        "reverbGainChanged",
        "reverbTimeChanged",
        "reverbBrightnessChanged",
        "updatePosition",
        "updateRotation",
        "dimensions",
        "QVector3D",
        "leftMaterial",
        "Material",
        "rightMaterial",
        "frontMaterial",
        "backMaterial",
        "floorMaterial",
        "ceilingMaterial",
        "reflectionGain",
        "reverbGain",
        "reverbTime",
        "reverbBrightness",
        "Transparent",
        "AcousticCeilingTiles",
        "BrickBare",
        "BrickPainted",
        "ConcreteBlockCoarse",
        "ConcreteBlockPainted",
        "CurtainHeavy",
        "FiberGlassInsulation",
        "GlassThin",
        "GlassThick",
        "Grass",
        "LinoleumOnConcrete",
        "Marble",
        "Metal",
        "ParquetOnConcrete",
        "PlasterRough",
        "PlasterSmooth",
        "PlywoodPanel",
        "PolishedConcreteOrTile",
        "Sheetrock",
        "WaterOrIceSurface",
        "WoodCeiling",
        "WoodPanel",
        "Uniform"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'dimensionsChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'rotationChanged'
        QtMocHelpers::SignalData<void()>(6, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'wallsChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reflectionGainChanged'
        QtMocHelpers::SignalData<void()>(8, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reverbGainChanged'
        QtMocHelpers::SignalData<void()>(9, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reverbTimeChanged'
        QtMocHelpers::SignalData<void()>(10, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reverbBrightnessChanged'
        QtMocHelpers::SignalData<void()>(11, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'updatePosition'
        QtMocHelpers::SlotData<void()>(12, 4, QMC::AccessProtected, QMetaType::Void),
        // Slot 'updateRotation'
        QtMocHelpers::SlotData<void()>(13, 4, QMC::AccessProtected, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'dimensions'
        QtMocHelpers::PropertyData<QVector3D>(14, 0x80000000 | 15, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1),
        // property 'leftMaterial'
        QtMocHelpers::PropertyData<enum Material>(16, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'rightMaterial'
        QtMocHelpers::PropertyData<enum Material>(18, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'frontMaterial'
        QtMocHelpers::PropertyData<enum Material>(19, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'backMaterial'
        QtMocHelpers::PropertyData<enum Material>(20, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'floorMaterial'
        QtMocHelpers::PropertyData<enum Material>(21, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'ceilingMaterial'
        QtMocHelpers::PropertyData<enum Material>(22, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 3),
        // property 'reflectionGain'
        QtMocHelpers::PropertyData<float>(23, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'reverbGain'
        QtMocHelpers::PropertyData<float>(24, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'reverbTime'
        QtMocHelpers::PropertyData<float>(25, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 6),
        // property 'reverbBrightness'
        QtMocHelpers::PropertyData<float>(26, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 7),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Material'
        QtMocHelpers::EnumData<enum Material>(17, 17, QMC::EnumFlags{}).add({
            {   27, Material::Transparent },
            {   28, Material::AcousticCeilingTiles },
            {   29, Material::BrickBare },
            {   30, Material::BrickPainted },
            {   31, Material::ConcreteBlockCoarse },
            {   32, Material::ConcreteBlockPainted },
            {   33, Material::CurtainHeavy },
            {   34, Material::FiberGlassInsulation },
            {   35, Material::GlassThin },
            {   36, Material::GlassThick },
            {   37, Material::Grass },
            {   38, Material::LinoleumOnConcrete },
            {   39, Material::Marble },
            {   40, Material::Metal },
            {   41, Material::ParquetOnConcrete },
            {   42, Material::PlasterRough },
            {   43, Material::PlasterSmooth },
            {   44, Material::PlywoodPanel },
            {   45, Material::PolishedConcreteOrTile },
            {   46, Material::Sheetrock },
            {   47, Material::WaterOrIceSurface },
            {   48, Material::WoodCeiling },
            {   49, Material::WoodPanel },
            {   50, Material::Uniform },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QQuick3DAudioRoom, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QQuick3DAudioRoom::staticMetaObject = { {
    QMetaObject::SuperData::link<QQuick3DNode::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QQuick3DAudioRoomE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QQuick3DAudioRoomE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17QQuick3DAudioRoomE_t>.metaTypes,
    nullptr
} };

void QQuick3DAudioRoom::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QQuick3DAudioRoom *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->positionChanged(); break;
        case 1: _t->dimensionsChanged(); break;
        case 2: _t->rotationChanged(); break;
        case 3: _t->wallsChanged(); break;
        case 4: _t->reflectionGainChanged(); break;
        case 5: _t->reverbGainChanged(); break;
        case 6: _t->reverbTimeChanged(); break;
        case 7: _t->reverbBrightnessChanged(); break;
        case 8: _t->updatePosition(); break;
        case 9: _t->updateRotation(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::positionChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::dimensionsChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::rotationChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::wallsChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::reflectionGainChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::reverbGainChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::reverbTimeChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioRoom::*)()>(_a, &QQuick3DAudioRoom::reverbBrightnessChanged, 7))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QVector3D*>(_v) = _t->dimensions(); break;
        case 1: *reinterpret_cast<enum Material*>(_v) = _t->leftMaterial(); break;
        case 2: *reinterpret_cast<enum Material*>(_v) = _t->rightMaterial(); break;
        case 3: *reinterpret_cast<enum Material*>(_v) = _t->frontMaterial(); break;
        case 4: *reinterpret_cast<enum Material*>(_v) = _t->backMaterial(); break;
        case 5: *reinterpret_cast<enum Material*>(_v) = _t->floorMaterial(); break;
        case 6: *reinterpret_cast<enum Material*>(_v) = _t->ceilingMaterial(); break;
        case 7: *reinterpret_cast<float*>(_v) = _t->reflectionGain(); break;
        case 8: *reinterpret_cast<float*>(_v) = _t->reverbGain(); break;
        case 9: *reinterpret_cast<float*>(_v) = _t->reverbTime(); break;
        case 10: *reinterpret_cast<float*>(_v) = _t->reverbBrightness(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setDimensions(*reinterpret_cast<QVector3D*>(_v)); break;
        case 1: _t->setLeftMaterial(*reinterpret_cast<enum Material*>(_v)); break;
        case 2: _t->setRightMaterial(*reinterpret_cast<enum Material*>(_v)); break;
        case 3: _t->setFrontMaterial(*reinterpret_cast<enum Material*>(_v)); break;
        case 4: _t->setBackMaterial(*reinterpret_cast<enum Material*>(_v)); break;
        case 5: _t->setFloorMaterial(*reinterpret_cast<enum Material*>(_v)); break;
        case 6: _t->setCeilingMaterial(*reinterpret_cast<enum Material*>(_v)); break;
        case 7: _t->setReflectionGain(*reinterpret_cast<float*>(_v)); break;
        case 8: _t->setReverbGain(*reinterpret_cast<float*>(_v)); break;
        case 9: _t->setReverbTime(*reinterpret_cast<float*>(_v)); break;
        case 10: _t->setReverbBrightness(*reinterpret_cast<float*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QQuick3DAudioRoom::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuick3DAudioRoom::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QQuick3DAudioRoomE_t>.strings))
        return static_cast<void*>(this);
    return QQuick3DNode::qt_metacast(_clname);
}

int QQuick3DAudioRoom::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QQuick3DNode::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 10)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 10;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 10)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 10;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    return _id;
}

// SIGNAL 0
void QQuick3DAudioRoom::positionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QQuick3DAudioRoom::dimensionsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QQuick3DAudioRoom::rotationChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QQuick3DAudioRoom::wallsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QQuick3DAudioRoom::reflectionGainChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QQuick3DAudioRoom::reverbGainChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QQuick3DAudioRoom::reverbTimeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QQuick3DAudioRoom::reverbBrightnessChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}
QT_WARNING_POP
