/****************************************************************************
** Meta object code from reading C++ file 'qquick3daudioengine_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/spatialaudioquick3d/qquick3daudioengine_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qquick3daudioengine_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN19QQuick3DAudioEngineE_t {};
} // unnamed namespace

template <> constexpr inline auto QQuick3DAudioEngine::qt_create_metaobjectdata<qt_meta_tag_ZN19QQuick3DAudioEngineE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QQuick3DAudioEngine",
        "QML.Element",
        "AudioEngine",
        "outputModeChanged",
        "",
        "outputDeviceChanged",
        "masterVolumeChanged",
        "distanceScaleChanged",
        "outputMode",
        "OutputMode",
        "outputDevice",
        "QAudioDevice",
        "masterVolume",
        "distanceScale",
        "Surround",
        "Stereo",
        "Headphone"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'outputModeChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'outputDeviceChanged'
        QtMocHelpers::SignalData<void()>(5, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'masterVolumeChanged'
        QtMocHelpers::SignalData<void()>(6, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'distanceScaleChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'outputMode'
        QtMocHelpers::PropertyData<enum OutputMode>(8, 0x80000000 | 9, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'outputDevice'
        QtMocHelpers::PropertyData<QAudioDevice>(10, 0x80000000 | 11, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1),
        // property 'masterVolume'
        QtMocHelpers::PropertyData<float>(12, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'distanceScale'
        QtMocHelpers::PropertyData<float>(13, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'OutputMode'
        QtMocHelpers::EnumData<enum OutputMode>(9, 9, QMC::EnumFlags{}).add({
            {   14, OutputMode::Surround },
            {   15, OutputMode::Stereo },
            {   16, OutputMode::Headphone },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QQuick3DAudioEngine, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QQuick3DAudioEngine::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QQuick3DAudioEngineE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QQuick3DAudioEngineE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QQuick3DAudioEngineE_t>.metaTypes,
    nullptr
} };

void QQuick3DAudioEngine::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QQuick3DAudioEngine *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->outputModeChanged(); break;
        case 1: _t->outputDeviceChanged(); break;
        case 2: _t->masterVolumeChanged(); break;
        case 3: _t->distanceScaleChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioEngine::*)()>(_a, &QQuick3DAudioEngine::outputModeChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioEngine::*)()>(_a, &QQuick3DAudioEngine::outputDeviceChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioEngine::*)()>(_a, &QQuick3DAudioEngine::masterVolumeChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuick3DAudioEngine::*)()>(_a, &QQuick3DAudioEngine::distanceScaleChanged, 3))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<enum OutputMode*>(_v) = _t->outputMode(); break;
        case 1: *reinterpret_cast<QAudioDevice*>(_v) = _t->outputDevice(); break;
        case 2: *reinterpret_cast<float*>(_v) = _t->masterVolume(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->distanceScale(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setOutputMode(*reinterpret_cast<enum OutputMode*>(_v)); break;
        case 1: _t->setOutputDevice(*reinterpret_cast<QAudioDevice*>(_v)); break;
        case 2: _t->setMasterVolume(*reinterpret_cast<float*>(_v)); break;
        case 3: _t->setDistanceScale(*reinterpret_cast<float*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QQuick3DAudioEngine::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuick3DAudioEngine::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QQuick3DAudioEngineE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QQuick3DAudioEngine::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 4;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void QQuick3DAudioEngine::outputModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QQuick3DAudioEngine::outputDeviceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QQuick3DAudioEngine::masterVolumeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QQuick3DAudioEngine::distanceScaleChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}
QT_WARNING_POP
