/****************************************************************************
** Meta object code from reading C++ file 'qquickvideooutput_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimediaquick/qquickvideooutput_p.h"
#include "qvideosink.h"
#include "qvideoframe.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qquickvideooutput_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN15QQuickVideoSinkE_t {};
} // unnamed namespace

template <> constexpr inline auto QQuickVideoSink::qt_create_metaobjectdata<qt_meta_tag_ZN15QQuickVideoSinkE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QQuickVideoSink",
        "QML.Element",
        "VideoSink",
        "videoFrameChanged",
        ""
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'videoFrameChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QQuickVideoSink, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QQuickVideoSink::staticMetaObject = { {
    QMetaObject::SuperData::link<QVideoSink::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15QQuickVideoSinkE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15QQuickVideoSinkE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN15QQuickVideoSinkE_t>.metaTypes,
    nullptr
} };

void QQuickVideoSink::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QQuickVideoSink *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->videoFrameChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoSink::*)()>(_a, &QQuickVideoSink::videoFrameChanged, 0))
            return;
    }
}

const QMetaObject *QQuickVideoSink::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuickVideoSink::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15QQuickVideoSinkE_t>.strings))
        return static_cast<void*>(this);
    return QVideoSink::qt_metacast(_clname);
}

int QQuickVideoSink::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QVideoSink::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}

// SIGNAL 0
void QQuickVideoSink::videoFrameChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}
namespace {
struct qt_meta_tag_ZN17QQuickVideoOutputE_t {};
} // unnamed namespace

template <> constexpr inline auto QQuickVideoOutput::qt_create_metaobjectdata<qt_meta_tag_ZN17QQuickVideoOutputE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QQuickVideoOutput",
        "QML.Element",
        "VideoOutput",
        "sourceChanged",
        "",
        "fillModeChanged",
        "QQuickVideoOutput::FillMode",
        "orientationChanged",
        "mirroredChanged",
        "sourceRectChanged",
        "contentRectChanged",
        "endOfStreamPolicyChanged",
        "QQuickVideoOutput::EndOfStreamPolicy",
        "_q_newFrame",
        "QSize",
        "_q_updateGeometry",
        "videoSink",
        "QVideoSink*",
        "clearOutput",
        "fillMode",
        "FillMode",
        "endOfStreamPolicy",
        "EndOfStreamPolicy",
        "orientation",
        "mirrored",
        "sourceRect",
        "QRectF",
        "contentRect",
        "Stretch",
        "PreserveAspectFit",
        "PreserveAspectCrop",
        "ClearOutput",
        "KeepLastFrame"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sourceChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'fillModeChanged'
        QtMocHelpers::SignalData<void(QQuickVideoOutput::FillMode)>(5, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 4 },
        }}),
        // Signal 'orientationChanged'
        QtMocHelpers::SignalData<void()>(7, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'mirroredChanged'
        QtMocHelpers::SignalData<void()>(8, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'sourceRectChanged'
        QtMocHelpers::SignalData<void()>(9, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'contentRectChanged'
        QtMocHelpers::SignalData<void()>(10, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'endOfStreamPolicyChanged'
        QtMocHelpers::SignalData<void(QQuickVideoOutput::EndOfStreamPolicy)>(11, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 4 },
        }}),
        // Slot '_q_newFrame'
        QtMocHelpers::SlotData<void(QSize)>(13, 4, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 14, 4 },
        }}),
        // Slot '_q_updateGeometry'
        QtMocHelpers::SlotData<void()>(15, 4, QMC::AccessPrivate, QMetaType::Void),
        // Method 'videoSink'
        QtMocHelpers::MethodData<QVideoSink *() const>(16, 4, QMC::AccessPublic, 0x80000000 | 17),
        // Method 'clearOutput'
        QtMocHelpers::RevisionedMethodData<void()>(18, 4, QMC::AccessPublic, 0x609, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'fillMode'
        QtMocHelpers::PropertyData<enum FillMode>(19, 0x80000000 | 20, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1),
        // property 'endOfStreamPolicy'
        QtMocHelpers::PropertyData<enum EndOfStreamPolicy>(21, 0x80000000 | 22, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 6, 0x609),
        // property 'orientation'
        QtMocHelpers::PropertyData<int>(23, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'mirrored'
        QtMocHelpers::PropertyData<bool>(24, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3, 0x609),
        // property 'sourceRect'
        QtMocHelpers::PropertyData<QRectF>(25, 0x80000000 | 26, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 4),
        // property 'contentRect'
        QtMocHelpers::PropertyData<QRectF>(27, 0x80000000 | 26, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 5),
        // property 'videoSink'
        QtMocHelpers::PropertyData<QVideoSink*>(16, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'FillMode'
        QtMocHelpers::EnumData<enum FillMode>(20, 20, QMC::EnumFlags{}).add({
            {   28, FillMode::Stretch },
            {   29, FillMode::PreserveAspectFit },
            {   30, FillMode::PreserveAspectCrop },
        }),
        // enum 'EndOfStreamPolicy'
        QtMocHelpers::EnumData<enum EndOfStreamPolicy>(22, 22, QMC::EnumFlags{}).add({
            {   31, EndOfStreamPolicy::ClearOutput },
            {   32, EndOfStreamPolicy::KeepLastFrame },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QQuickVideoOutput, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QQuickVideoOutput::staticMetaObject = { {
    QMetaObject::SuperData::link<QQuickItem::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QQuickVideoOutputE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QQuickVideoOutputE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17QQuickVideoOutputE_t>.metaTypes,
    nullptr
} };

void QQuickVideoOutput::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QQuickVideoOutput *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sourceChanged(); break;
        case 1: _t->fillModeChanged((*reinterpret_cast<std::add_pointer_t<QQuickVideoOutput::FillMode>>(_a[1]))); break;
        case 2: _t->orientationChanged(); break;
        case 3: _t->mirroredChanged(); break;
        case 4: _t->sourceRectChanged(); break;
        case 5: _t->contentRectChanged(); break;
        case 6: _t->endOfStreamPolicyChanged((*reinterpret_cast<std::add_pointer_t<QQuickVideoOutput::EndOfStreamPolicy>>(_a[1]))); break;
        case 7: _t->_q_newFrame((*reinterpret_cast<std::add_pointer_t<QSize>>(_a[1]))); break;
        case 8: _t->_q_updateGeometry(); break;
        case 9: { QVideoSink* _r = _t->videoSink();
            if (_a[0]) *reinterpret_cast<QVideoSink**>(_a[0]) = std::move(_r); }  break;
        case 10: _t->clearOutput(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)()>(_a, &QQuickVideoOutput::sourceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)(QQuickVideoOutput::FillMode )>(_a, &QQuickVideoOutput::fillModeChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)()>(_a, &QQuickVideoOutput::orientationChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)()>(_a, &QQuickVideoOutput::mirroredChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)()>(_a, &QQuickVideoOutput::sourceRectChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)()>(_a, &QQuickVideoOutput::contentRectChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QQuickVideoOutput::*)(QQuickVideoOutput::EndOfStreamPolicy )>(_a, &QQuickVideoOutput::endOfStreamPolicyChanged, 6))
            return;
    }
    if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 6:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QVideoSink* >(); break;
        }
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<enum FillMode*>(_v) = _t->fillMode(); break;
        case 1: *reinterpret_cast<enum EndOfStreamPolicy*>(_v) = _t->endOfStreamPolicy(); break;
        case 2: *reinterpret_cast<int*>(_v) = _t->orientation(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->mirrored(); break;
        case 4: *reinterpret_cast<QRectF*>(_v) = _t->sourceRect(); break;
        case 5: *reinterpret_cast<QRectF*>(_v) = _t->contentRect(); break;
        case 6: *reinterpret_cast<QVideoSink**>(_v) = _t->videoSink(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setFillMode(*reinterpret_cast<enum FillMode*>(_v)); break;
        case 1: _t->setEndOfStreamPolicy(*reinterpret_cast<enum EndOfStreamPolicy*>(_v)); break;
        case 2: _t->setOrientation(*reinterpret_cast<int*>(_v)); break;
        case 3: _t->setMirrored(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QQuickVideoOutput::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QQuickVideoOutput::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17QQuickVideoOutputE_t>.strings))
        return static_cast<void*>(this);
    return QQuickItem::qt_metacast(_clname);
}

int QQuickVideoOutput::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QQuickItem::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 11;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void QQuickVideoOutput::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QQuickVideoOutput::fillModeChanged(QQuickVideoOutput::FillMode _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QQuickVideoOutput::orientationChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QQuickVideoOutput::mirroredChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QQuickVideoOutput::sourceRectChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QQuickVideoOutput::contentRectChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QQuickVideoOutput::endOfStreamPolicyChanged(QQuickVideoOutput::EndOfStreamPolicy _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}
QT_WARNING_POP
