/****************************************************************************
** Meta object code from reading C++ file 'qtmultimediaquicktypes_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimediaquick/qtmultimediaquicktypes_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qtmultimediaquicktypes_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate27QMediaCaptureSessionForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaCaptureSessionForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate27QMediaCaptureSessionForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaCaptureSessionForeign",
        "QML.Foreign",
        "QMediaCaptureSession",
        "QML.Element",
        "CaptureSession"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QMediaCaptureSessionForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaCaptureSessionForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate27QMediaCaptureSessionForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate27QMediaCaptureSessionForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate27QMediaCaptureSessionForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate14QCameraForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QCameraForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate14QCameraForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QCameraForeign",
        "QML.Foreign",
        "QCamera",
        "QML.Element",
        "Camera"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QCameraForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QCameraForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate14QCameraForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate14QCameraForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate14QCameraForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate20QImageCaptureForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QImageCaptureForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate20QImageCaptureForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QImageCaptureForeign",
        "QML.Element",
        "anonymous",
        "QML.Foreign",
        "QImageCapture"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QImageCaptureForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QImageCaptureForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QImageCaptureForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QImageCaptureForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate20QImageCaptureForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate21QScreenCaptureForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QScreenCaptureForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate21QScreenCaptureForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QScreenCaptureForeign",
        "QML.Element",
        "anonymous",
        "QML.Foreign",
        "QScreenCapture"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QScreenCaptureForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QScreenCaptureForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QScreenCaptureForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QScreenCaptureForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate21QScreenCaptureForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate21QMediaRecorderForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaRecorderForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaRecorderForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaRecorderForeign",
        "QML.Foreign",
        "QMediaRecorder",
        "QML.Element",
        "MediaRecorder"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QMediaRecorderForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaRecorderForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaRecorderForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaRecorderForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaRecorderForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaMetaDataForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaMetaDataForeign",
        "QML.Foreign",
        "QMediaMetaData",
        "QML.Element",
        "mediaMetaData"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QMediaMetaDataForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaMetaDataForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataDerivedE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaMetaDataDerived::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataDerivedE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaMetaDataDerived"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QMediaMetaDataDerived, qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataDerivedE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaMetaDataDerived::staticMetaObject = { {
    QtPrivate::MetaObjectForType<QMediaMetaData>::value,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataDerivedE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataDerivedE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate21QMediaMetaDataDerivedE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate20QMediaDevicesForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaDevicesForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate20QMediaDevicesForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaDevicesForeign",
        "QML.Foreign",
        "QMediaDevices",
        "QML.Element",
        "MediaDevices"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QMediaDevicesForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaDevicesForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QMediaDevicesForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QMediaDevicesForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate20QMediaDevicesForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate18QAudioInputForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QAudioInputForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate18QAudioInputForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QAudioInputForeign",
        "QML.Foreign",
        "QAudioInput",
        "QML.Element",
        "AudioInput"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QAudioInputForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QAudioInputForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate18QAudioInputForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate18QAudioInputForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate18QAudioInputForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate19QAudioOutputForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QAudioOutputForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioOutputForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QAudioOutputForeign",
        "QML.Foreign",
        "QAudioOutput",
        "QML.Element",
        "AudioOutput"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QAudioOutputForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QAudioOutputForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioOutputForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioOutputForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioOutputForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QAudioDeviceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QAudioDeviceForeign",
        "QML.Foreign",
        "QAudioDevice",
        "QML.Element",
        "audioDevice"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QAudioDeviceForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QAudioDeviceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceDerivedE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QAudioDeviceDerived::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceDerivedE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QAudioDeviceDerived"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QAudioDeviceDerived, qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceDerivedE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QAudioDeviceDerived::staticMetaObject = { {
    QtPrivate::MetaObjectForType<QAudioDevice>::value,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceDerivedE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceDerivedE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate19QAudioDeviceDerivedE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QCameraDeviceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QCameraDeviceForeign",
        "QML.Foreign",
        "QCameraDevice",
        "QML.Element",
        "cameraDevice"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QCameraDeviceForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QCameraDeviceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceDerivedE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QCameraDeviceDerived::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceDerivedE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QCameraDeviceDerived"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QCameraDeviceDerived, qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceDerivedE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QCameraDeviceDerived::staticMetaObject = { {
    QtPrivate::MetaObjectForType<QCameraDevice>::value,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceDerivedE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceDerivedE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraDeviceDerivedE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaFormatForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaFormatForeign",
        "QML.Foreign",
        "QMediaFormat",
        "QML.Element",
        "mediaFormat"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QMediaFormatForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaFormatForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatDerivedE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaFormatDerived::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatDerivedE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaFormatDerived"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QMediaFormatDerived, qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatDerivedE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaFormatDerived::staticMetaObject = { {
    QtPrivate::MetaObjectForType<QMediaFormat>::value,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatDerivedE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatDerivedE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate19QMediaFormatDerivedE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate20QCameraFormatForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QCameraFormatForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraFormatForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QCameraFormatForeign",
        "QML.Foreign",
        "QCameraFormat",
        "QML.Element",
        "cameraFormat"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QCameraFormatForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QCameraFormatForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraFormatForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraFormatForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate20QCameraFormatForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate24QCapturableWindowForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QCapturableWindowForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate24QCapturableWindowForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QCapturableWindowForeign",
        "QML.Foreign",
        "QCapturableWindow",
        "QML.Element",
        "capturableWindow",
        "QML.Creatable",
        "true",
        "QML.CreationMethod",
        "construct"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
            {    7,    8 },
    });
    return QtMocHelpers::metaObjectData<QCapturableWindowForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QCapturableWindowForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate24QCapturableWindowForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate24QCapturableWindowForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate24QCapturableWindowForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate21QWindowCaptureForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QWindowCaptureForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate21QWindowCaptureForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QWindowCaptureForeign",
        "QML.Foreign",
        "QWindowCapture",
        "QML.Element",
        "WindowCapture"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<QWindowCaptureForeign, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QWindowCaptureForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QWindowCaptureForeignE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate21QWindowCaptureForeignE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate21QWindowCaptureForeignE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate23QPlaybackOptionsDerivedE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QPlaybackOptionsDerived::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate23QPlaybackOptionsDerivedE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QPlaybackOptionsDerived",
        "QML.Foreign",
        "QPlaybackOptions",
        "QML.Element",
        "playbackOptions",
        "QML.Extended",
        "QPlaybackOptionsDerived",
        "QML.AddedInVersion",
        "1546",
        "networkTimeoutMs"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
        // property 'networkTimeoutMs'
        QtMocHelpers::PropertyData<qint64>(9, QMetaType::LongLong, QMC::DefaultPropertyFlags | QMC::Writable | QMC::Resettable | QMC::StdCppSet | QMC::Final),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
            {    7,    8 },
    });
    return QtMocHelpers::metaObjectData<QPlaybackOptionsDerived, void>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QPlaybackOptionsDerived::staticMetaObject = { {
    QtPrivate::MetaObjectForType<QPlaybackOptions>::value,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate23QPlaybackOptionsDerivedE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19QtMultimediaPrivate23QPlaybackOptionsDerivedE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19QtMultimediaPrivate23QPlaybackOptionsDerivedE_t>.metaTypes,
    nullptr
} };

void QtMultimediaPrivate::QPlaybackOptionsDerived::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = reinterpret_cast<QPlaybackOptionsDerived *>(_o);
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<qint64*>(_v) = _t->networkTimeoutMs(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setNetworkTimeoutMs(*reinterpret_cast<qint64*>(_v)); break;
        default: break;
        }
    }
    if (_c == QMetaObject::ResetProperty) {
        switch (_id) {
        case 0: _t->resetNetworkTimeoutMs(); break;
        default: break;
        }
    }
}
namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaMetaDataNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaMetaDataNamespaceForeign",
        "QML.Foreign",
        "QMediaMetaDataDerived",
        "QML.ForeignIsNamespace",
        "true",
        "QML.Element",
        "MediaMetaData"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
    });
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}

static constexpr auto qt_staticMetaObjectContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE =
    QtMultimediaPrivate::QMediaMetaDataNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE.relocatingData;

Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaMetaDataNamespaceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE.stringdata,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate30QMediaMetaDataNamespaceForeignE.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QAudioDeviceNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QAudioDeviceNamespaceForeign",
        "QML.Foreign",
        "QAudioDeviceDerived",
        "QML.ForeignIsNamespace",
        "true",
        "QML.Element",
        "AudioDevice"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
    });
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}

static constexpr auto qt_staticMetaObjectContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE =
    QtMultimediaPrivate::QAudioDeviceNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE.relocatingData;

Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QAudioDeviceNamespaceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE.stringdata,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate28QAudioDeviceNamespaceForeignE.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QCameraDeviceNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QCameraDeviceNamespaceForeign",
        "QML.Foreign",
        "QCameraDeviceDerived",
        "QML.ForeignIsNamespace",
        "true",
        "QML.Element",
        "CameraDevice"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
    });
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}

static constexpr auto qt_staticMetaObjectContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE =
    QtMultimediaPrivate::QCameraDeviceNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE.relocatingData;

Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QCameraDeviceNamespaceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE.stringdata,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate29QCameraDeviceNamespaceForeignE.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QMediaFormatNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QMediaFormatNamespaceForeign",
        "QML.Foreign",
        "QMediaFormatDerived",
        "QML.ForeignIsNamespace",
        "true",
        "QML.Element",
        "MediaFormat"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
    });
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}

static constexpr auto qt_staticMetaObjectContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE =
    QtMultimediaPrivate::QMediaFormatNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE.relocatingData;

Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QMediaFormatNamespaceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE.stringdata,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate28QMediaFormatNamespaceForeignE.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE_t {};
} // unnamed namespace

template <> constexpr inline auto QtMultimediaPrivate::QPlaybackOptionsNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QtMultimediaPrivate::QPlaybackOptionsNamespaceForeign",
        "QML.Element",
        "PlaybackOptions",
        "QML.Foreign",
        "QPlaybackOptions",
        "QML.ForeignIsNamespace",
        "true",
        "QML.AddedInVersion",
        "1546"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
            {    5,    6 },
            {    7,    8 },
    });
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}

static constexpr auto qt_staticMetaObjectContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE =
    QtMultimediaPrivate::QPlaybackOptionsNamespaceForeign::qt_create_metaobjectdata<qt_meta_tag_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE =
    qt_staticMetaObjectContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE.relocatingData;

Q_CONSTINIT const QMetaObject QtMultimediaPrivate::QPlaybackOptionsNamespaceForeign::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE.stringdata,
    qt_staticMetaObjectStaticContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN19QtMultimediaPrivate32QPlaybackOptionsNamespaceForeignE.metaTypes,
    nullptr
} };

QT_WARNING_POP
