/****************************************************************************
** Meta object code from reading C++ file 'qaudiodecoder.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/audio/qaudiodecoder.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qaudiodecoder.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QAudioDecoderE_t {};
} // unnamed namespace

template <> constexpr inline auto QAudioDecoder::qt_create_metaobjectdata<qt_meta_tag_ZN13QAudioDecoderE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QAudioDecoder",
        "bufferAvailableChanged",
        "",
        "bufferReady",
        "finished",
        "isDecodingChanged",
        "formatChanged",
        "QAudioFormat",
        "format",
        "error",
        "QAudioDecoder::Error",
        "sourceChanged",
        "positionChanged",
        "position",
        "durationChanged",
        "duration",
        "start",
        "stop",
        "source",
        "QUrl",
        "isDecoding",
        "bufferAvailable",
        "Error",
        "NoError",
        "ResourceError",
        "FormatError",
        "AccessDeniedError",
        "NotSupportedError"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'bufferAvailableChanged'
        QtMocHelpers::SignalData<void(bool)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'bufferReady'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'finished'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'isDecodingChanged'
        QtMocHelpers::SignalData<void(bool)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'formatChanged'
        QtMocHelpers::SignalData<void(const QAudioFormat &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'error'
        QtMocHelpers::SignalData<void(QAudioDecoder::Error)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 9 },
        }}),
        // Signal 'sourceChanged'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void(qint64)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 13 },
        }}),
        // Signal 'durationChanged'
        QtMocHelpers::SignalData<void(qint64)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 15 },
        }}),
        // Slot 'start'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'source'
        QtMocHelpers::PropertyData<QUrl>(18, 0x80000000 | 19, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 6),
        // property 'isDecoding'
        QtMocHelpers::PropertyData<bool>(20, QMetaType::Bool, QMC::DefaultPropertyFlags, 3),
        // property 'error'
        QtMocHelpers::PropertyData<QString>(9, QMetaType::QString, QMC::DefaultPropertyFlags),
        // property 'bufferAvailable'
        QtMocHelpers::PropertyData<bool>(21, QMetaType::Bool, QMC::DefaultPropertyFlags, 0),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Error'
        QtMocHelpers::EnumData<enum Error>(22, 22, QMC::EnumFlags{}).add({
            {   23, Error::NoError },
            {   24, Error::ResourceError },
            {   25, Error::FormatError },
            {   26, Error::AccessDeniedError },
            {   27, Error::NotSupportedError },
        }),
    };
    return QtMocHelpers::metaObjectData<QAudioDecoder, qt_meta_tag_ZN13QAudioDecoderE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QAudioDecoder::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QAudioDecoderE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QAudioDecoderE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QAudioDecoderE_t>.metaTypes,
    nullptr
} };

void QAudioDecoder::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QAudioDecoder *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->bufferAvailableChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 1: _t->bufferReady(); break;
        case 2: _t->finished(); break;
        case 3: _t->isDecodingChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 4: _t->formatChanged((*reinterpret_cast<std::add_pointer_t<QAudioFormat>>(_a[1]))); break;
        case 5: _t->error((*reinterpret_cast<std::add_pointer_t<QAudioDecoder::Error>>(_a[1]))); break;
        case 6: _t->sourceChanged(); break;
        case 7: _t->positionChanged((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 8: _t->durationChanged((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 9: _t->start(); break;
        case 10: _t->stop(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 4:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QAudioFormat >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)(bool )>(_a, &QAudioDecoder::bufferAvailableChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)()>(_a, &QAudioDecoder::bufferReady, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)()>(_a, &QAudioDecoder::finished, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)(bool )>(_a, &QAudioDecoder::isDecodingChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)(const QAudioFormat & )>(_a, &QAudioDecoder::formatChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)(QAudioDecoder::Error )>(_a, &QAudioDecoder::error, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)()>(_a, &QAudioDecoder::sourceChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)(qint64 )>(_a, &QAudioDecoder::positionChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioDecoder::*)(qint64 )>(_a, &QAudioDecoder::durationChanged, 8))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QUrl*>(_v) = _t->source(); break;
        case 1: *reinterpret_cast<bool*>(_v) = _t->isDecoding(); break;
        case 2: *reinterpret_cast<QString*>(_v) = _t->errorString(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->bufferAvailable(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setSource(*reinterpret_cast<QUrl*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QAudioDecoder::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QAudioDecoder::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QAudioDecoderE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QAudioDecoder::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void QAudioDecoder::bufferAvailableChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QAudioDecoder::bufferReady()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QAudioDecoder::finished()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QAudioDecoder::isDecodingChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void QAudioDecoder::formatChanged(const QAudioFormat & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void QAudioDecoder::error(QAudioDecoder::Error _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void QAudioDecoder::sourceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QAudioDecoder::positionChanged(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void QAudioDecoder::durationChanged(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}
QT_WARNING_POP
