// This file is autogenerated. Changes will be overwritten.
#include "EWIEGA46WW/moc_qquick3dambientsound_p.cpp"
#include "EWIEGA46WW/moc_qquick3daudiolistener_p.cpp"
#include "EWIEGA46WW/moc_qquick3daudioroom_p.cpp"
#include "EWIEGA46WW/moc_qquick3dspatialsound_p.cpp"
