/****************************************************************************
** Meta object code from reading C++ file 'qplatformimagecapture_p.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/platform/qplatformimagecapture_p.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qplatformimagecapture_p.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN21QPlatformImageCaptureE_t {};
} // unnamed namespace

template <> constexpr inline auto QPlatformImageCapture::qt_create_metaobjectdata<qt_meta_tag_ZN21QPlatformImageCaptureE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QPlatformImageCapture",
        "readyForCaptureChanged",
        "",
        "ready",
        "imageExposed",
        "requestId",
        "imageCaptured",
        "QImage",
        "preview",
        "imageMetadataAvailable",
        "id",
        "QMediaMetaData",
        "imageAvailable",
        "QVideoFrame",
        "buffer",
        "imageSaved",
        "fileName",
        "error",
        "errorString"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'readyForCaptureChanged'
        QtMocHelpers::SignalData<void(bool)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 3 },
        }}),
        // Signal 'imageExposed'
        QtMocHelpers::SignalData<void(int)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 },
        }}),
        // Signal 'imageCaptured'
        QtMocHelpers::SignalData<void(int, const QImage &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 }, { 0x80000000 | 7, 8 },
        }}),
        // Signal 'imageMetadataAvailable'
        QtMocHelpers::SignalData<void(int, const QMediaMetaData &)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 }, { 0x80000000 | 11, 2 },
        }}),
        // Signal 'imageAvailable'
        QtMocHelpers::SignalData<void(int, const QVideoFrame &)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 }, { 0x80000000 | 13, 14 },
        }}),
        // Signal 'imageSaved'
        QtMocHelpers::SignalData<void(int, const QString &)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 }, { QMetaType::QString, 16 },
        }}),
        // Signal 'error'
        QtMocHelpers::SignalData<void(int, int, const QString &)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 }, { QMetaType::Int, 17 }, { QMetaType::QString, 18 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QPlatformImageCapture, qt_meta_tag_ZN21QPlatformImageCaptureE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QPlatformImageCapture::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN21QPlatformImageCaptureE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN21QPlatformImageCaptureE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN21QPlatformImageCaptureE_t>.metaTypes,
    nullptr
} };

void QPlatformImageCapture::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QPlatformImageCapture *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->readyForCaptureChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 1: _t->imageExposed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 2: _t->imageCaptured((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QImage>>(_a[2]))); break;
        case 3: _t->imageMetadataAvailable((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QMediaMetaData>>(_a[2]))); break;
        case 4: _t->imageAvailable((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVideoFrame>>(_a[2]))); break;
        case 5: _t->imageSaved((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 6: _t->error((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 3:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QMediaMetaData >(); break;
            }
            break;
        case 4:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QVideoFrame >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(bool )>(_a, &QPlatformImageCapture::readyForCaptureChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(int )>(_a, &QPlatformImageCapture::imageExposed, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(int , const QImage & )>(_a, &QPlatformImageCapture::imageCaptured, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(int , const QMediaMetaData & )>(_a, &QPlatformImageCapture::imageMetadataAvailable, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(int , const QVideoFrame & )>(_a, &QPlatformImageCapture::imageAvailable, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(int , const QString & )>(_a, &QPlatformImageCapture::imageSaved, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QPlatformImageCapture::*)(int , int , const QString & )>(_a, &QPlatformImageCapture::error, 6))
            return;
    }
}

const QMetaObject *QPlatformImageCapture::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QPlatformImageCapture::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN21QPlatformImageCaptureE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QPlatformImageCapture::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 7)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void QPlatformImageCapture::readyForCaptureChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QPlatformImageCapture::imageExposed(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QPlatformImageCapture::imageCaptured(int _t1, const QImage & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void QPlatformImageCapture::imageMetadataAvailable(int _t1, const QMediaMetaData & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2);
}

// SIGNAL 4
void QPlatformImageCapture::imageAvailable(int _t1, const QVideoFrame & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2);
}

// SIGNAL 5
void QPlatformImageCapture::imageSaved(int _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2);
}

// SIGNAL 6
void QPlatformImageCapture::error(int _t1, int _t2, const QString & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2, _t3);
}
QT_WARNING_POP
