/****************************************************************************
** Meta object code from reading C++ file 'qmediarecorder.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/recording/qmediarecorder.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qmediarecorder.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN14QMediaRecorderE_t {};
} // unnamed namespace

template <> constexpr inline auto QMediaRecorder::qt_create_metaobjectdata<qt_meta_tag_ZN14QMediaRecorderE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QMediaRecorder",
        "recorderStateChanged",
        "",
        "QMediaRecorder::RecorderState",
        "state",
        "durationChanged",
        "duration",
        "actualLocationChanged",
        "QUrl",
        "location",
        "encoderSettingsChanged",
        "errorOccurred",
        "QMediaRecorder::Error",
        "error",
        "errorString",
        "errorChanged",
        "metaDataChanged",
        "mediaFormatChanged",
        "encodingModeChanged",
        "qualityChanged",
        "videoResolutionChanged",
        "videoFrameRateChanged",
        "videoBitRateChanged",
        "audioBitRateChanged",
        "audioChannelCountChanged",
        "audioSampleRateChanged",
        "autoStopChanged",
        "record",
        "pause",
        "stop",
        "recorderState",
        "outputLocation",
        "actualLocation",
        "metaData",
        "QMediaMetaData",
        "mediaFormat",
        "QMediaFormat",
        "quality",
        "Quality",
        "encodingMode",
        "QMediaRecorder::EncodingMode",
        "videoResolution",
        "QSize",
        "videoFrameRate",
        "videoBitRate",
        "audioBitRate",
        "audioChannelCount",
        "audioSampleRate",
        "autoStop",
        "VeryLowQuality",
        "LowQuality",
        "NormalQuality",
        "HighQuality",
        "VeryHighQuality",
        "EncodingMode",
        "ConstantQualityEncoding",
        "ConstantBitRateEncoding",
        "AverageBitRateEncoding",
        "TwoPassEncoding",
        "RecorderState",
        "StoppedState",
        "RecordingState",
        "PausedState",
        "Error",
        "NoError",
        "ResourceError",
        "FormatError",
        "OutOfSpaceError",
        "LocationNotWritable"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'recorderStateChanged'
        QtMocHelpers::SignalData<void(QMediaRecorder::RecorderState)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'durationChanged'
        QtMocHelpers::SignalData<void(qint64)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 6 },
        }}),
        // Signal 'actualLocationChanged'
        QtMocHelpers::SignalData<void(const QUrl &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
        // Signal 'encoderSettingsChanged'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic | QMC::MethodCompatibility, QMetaType::Void),
        // Signal 'errorOccurred'
        QtMocHelpers::SignalData<void(QMediaRecorder::Error, const QString &)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 13 }, { QMetaType::QString, 14 },
        }}),
        // Signal 'errorChanged'
        QtMocHelpers::SignalData<void()>(15, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'metaDataChanged'
        QtMocHelpers::SignalData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'mediaFormatChanged'
        QtMocHelpers::SignalData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'encodingModeChanged'
        QtMocHelpers::SignalData<void()>(18, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'qualityChanged'
        QtMocHelpers::SignalData<void()>(19, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoResolutionChanged'
        QtMocHelpers::SignalData<void()>(20, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoFrameRateChanged'
        QtMocHelpers::SignalData<void()>(21, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'videoBitRateChanged'
        QtMocHelpers::SignalData<void()>(22, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioBitRateChanged'
        QtMocHelpers::SignalData<void()>(23, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioChannelCountChanged'
        QtMocHelpers::SignalData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioSampleRateChanged'
        QtMocHelpers::SignalData<void()>(25, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'autoStopChanged'
        QtMocHelpers::RevisionedSignalData<void()>(26, 2, QMC::AccessPublic, 0x608, QMetaType::Void),
        // Slot 'record'
        QtMocHelpers::SlotData<void()>(27, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'pause'
        QtMocHelpers::SlotData<void()>(28, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(29, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'recorderState'
        QtMocHelpers::PropertyData<QMediaRecorder::RecorderState>(30, 0x80000000 | 3, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 0),
        // property 'duration'
        QtMocHelpers::PropertyData<qint64>(6, QMetaType::LongLong, QMC::DefaultPropertyFlags, 1),
        // property 'outputLocation'
        QtMocHelpers::PropertyData<QUrl>(31, 0x80000000 | 8, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet),
        // property 'actualLocation'
        QtMocHelpers::PropertyData<QUrl>(32, 0x80000000 | 8, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 2),
        // property 'metaData'
        QtMocHelpers::PropertyData<QMediaMetaData>(33, 0x80000000 | 34, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 6),
        // property 'error'
        QtMocHelpers::PropertyData<QMediaRecorder::Error>(13, 0x80000000 | 12, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 5),
        // property 'errorString'
        QtMocHelpers::PropertyData<QString>(14, QMetaType::QString, QMC::DefaultPropertyFlags, 5),
        // property 'mediaFormat'
        QtMocHelpers::PropertyData<QMediaFormat>(35, 0x80000000 | 36, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 7),
        // property 'quality'
        QtMocHelpers::PropertyData<enum Quality>(37, 0x80000000 | 38, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 9),
        // property 'encodingMode'
        QtMocHelpers::PropertyData<QMediaRecorder::EncodingMode>(39, 0x80000000 | 40, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 8),
        // property 'videoResolution'
        QtMocHelpers::PropertyData<QSize>(41, 0x80000000 | 42, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 10),
        // property 'videoFrameRate'
        QtMocHelpers::PropertyData<qreal>(43, QMetaType::QReal, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 11),
        // property 'videoBitRate'
        QtMocHelpers::PropertyData<int>(44, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 12),
        // property 'audioBitRate'
        QtMocHelpers::PropertyData<int>(45, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 13),
        // property 'audioChannelCount'
        QtMocHelpers::PropertyData<int>(46, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 14),
        // property 'audioSampleRate'
        QtMocHelpers::PropertyData<int>(47, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 15),
        // property 'autoStop'
        QtMocHelpers::PropertyData<bool>(48, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 16, 0x608),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Quality'
        QtMocHelpers::EnumData<enum Quality>(38, 38, QMC::EnumFlags{}).add({
            {   49, Quality::VeryLowQuality },
            {   50, Quality::LowQuality },
            {   51, Quality::NormalQuality },
            {   52, Quality::HighQuality },
            {   53, Quality::VeryHighQuality },
        }),
        // enum 'EncodingMode'
        QtMocHelpers::EnumData<enum EncodingMode>(54, 54, QMC::EnumFlags{}).add({
            {   55, EncodingMode::ConstantQualityEncoding },
            {   56, EncodingMode::ConstantBitRateEncoding },
            {   57, EncodingMode::AverageBitRateEncoding },
            {   58, EncodingMode::TwoPassEncoding },
        }),
        // enum 'RecorderState'
        QtMocHelpers::EnumData<enum RecorderState>(59, 59, QMC::EnumFlags{}).add({
            {   60, RecorderState::StoppedState },
            {   61, RecorderState::RecordingState },
            {   62, RecorderState::PausedState },
        }),
        // enum 'Error'
        QtMocHelpers::EnumData<enum Error>(63, 63, QMC::EnumFlags{}).add({
            {   64, Error::NoError },
            {   65, Error::ResourceError },
            {   66, Error::FormatError },
            {   67, Error::OutOfSpaceError },
            {   68, Error::LocationNotWritable },
        }),
    };
    return QtMocHelpers::metaObjectData<QMediaRecorder, qt_meta_tag_ZN14QMediaRecorderE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QMediaRecorder::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14QMediaRecorderE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14QMediaRecorderE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN14QMediaRecorderE_t>.metaTypes,
    nullptr
} };

void QMediaRecorder::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QMediaRecorder *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->recorderStateChanged((*reinterpret_cast<std::add_pointer_t<QMediaRecorder::RecorderState>>(_a[1]))); break;
        case 1: _t->durationChanged((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 2: _t->actualLocationChanged((*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[1]))); break;
        case 3: _t->encoderSettingsChanged(); break;
        case 4: _t->errorOccurred((*reinterpret_cast<std::add_pointer_t<QMediaRecorder::Error>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 5: _t->errorChanged(); break;
        case 6: _t->metaDataChanged(); break;
        case 7: _t->mediaFormatChanged(); break;
        case 8: _t->encodingModeChanged(); break;
        case 9: _t->qualityChanged(); break;
        case 10: _t->videoResolutionChanged(); break;
        case 11: _t->videoFrameRateChanged(); break;
        case 12: _t->videoBitRateChanged(); break;
        case 13: _t->audioBitRateChanged(); break;
        case 14: _t->audioChannelCountChanged(); break;
        case 15: _t->audioSampleRateChanged(); break;
        case 16: _t->autoStopChanged(); break;
        case 17: _t->record(); break;
        case 18: _t->pause(); break;
        case 19: _t->stop(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)(QMediaRecorder::RecorderState )>(_a, &QMediaRecorder::recorderStateChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)(qint64 )>(_a, &QMediaRecorder::durationChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)(const QUrl & )>(_a, &QMediaRecorder::actualLocationChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::encoderSettingsChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)(QMediaRecorder::Error , const QString & )>(_a, &QMediaRecorder::errorOccurred, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::errorChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::metaDataChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::mediaFormatChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::encodingModeChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::qualityChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::videoResolutionChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::videoFrameRateChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::videoBitRateChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::audioBitRateChanged, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::audioChannelCountChanged, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::audioSampleRateChanged, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (QMediaRecorder::*)()>(_a, &QMediaRecorder::autoStopChanged, 16))
            return;
    }
    if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 4:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QMediaMetaData >(); break;
        }
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QMediaRecorder::RecorderState*>(_v) = _t->recorderState(); break;
        case 1: *reinterpret_cast<qint64*>(_v) = _t->duration(); break;
        case 2: *reinterpret_cast<QUrl*>(_v) = _t->outputLocation(); break;
        case 3: *reinterpret_cast<QUrl*>(_v) = _t->actualLocation(); break;
        case 4: *reinterpret_cast<QMediaMetaData*>(_v) = _t->metaData(); break;
        case 5: *reinterpret_cast<QMediaRecorder::Error*>(_v) = _t->error(); break;
        case 6: *reinterpret_cast<QString*>(_v) = _t->errorString(); break;
        case 7: *reinterpret_cast<QMediaFormat*>(_v) = _t->mediaFormat(); break;
        case 8: *reinterpret_cast<enum Quality*>(_v) = _t->quality(); break;
        case 9: *reinterpret_cast<QMediaRecorder::EncodingMode*>(_v) = _t->encodingMode(); break;
        case 10: *reinterpret_cast<QSize*>(_v) = _t->videoResolution(); break;
        case 11: *reinterpret_cast<qreal*>(_v) = _t->videoFrameRate(); break;
        case 12: *reinterpret_cast<int*>(_v) = _t->videoBitRate(); break;
        case 13: *reinterpret_cast<int*>(_v) = _t->audioBitRate(); break;
        case 14: *reinterpret_cast<int*>(_v) = _t->audioChannelCount(); break;
        case 15: *reinterpret_cast<int*>(_v) = _t->audioSampleRate(); break;
        case 16: *reinterpret_cast<bool*>(_v) = _t->autoStop(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 2: _t->setOutputLocation(*reinterpret_cast<QUrl*>(_v)); break;
        case 4: _t->setMetaData(*reinterpret_cast<QMediaMetaData*>(_v)); break;
        case 7: _t->setMediaFormat(*reinterpret_cast<QMediaFormat*>(_v)); break;
        case 8: _t->setQuality(*reinterpret_cast<enum Quality*>(_v)); break;
        case 9: _t->setEncodingMode(*reinterpret_cast<QMediaRecorder::EncodingMode*>(_v)); break;
        case 10: _t->setVideoResolution(*reinterpret_cast<QSize*>(_v)); break;
        case 11: _t->setVideoFrameRate(*reinterpret_cast<qreal*>(_v)); break;
        case 12: _t->setVideoBitRate(*reinterpret_cast<int*>(_v)); break;
        case 13: _t->setAudioBitRate(*reinterpret_cast<int*>(_v)); break;
        case 14: _t->setAudioChannelCount(*reinterpret_cast<int*>(_v)); break;
        case 15: _t->setAudioSampleRate(*reinterpret_cast<int*>(_v)); break;
        case 16: _t->setAutoStop(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QMediaRecorder::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QMediaRecorder::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14QMediaRecorderE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QMediaRecorder::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 20)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 20;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 20)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 20;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 17;
    }
    return _id;
}

// SIGNAL 0
void QMediaRecorder::recorderStateChanged(QMediaRecorder::RecorderState _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QMediaRecorder::durationChanged(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void QMediaRecorder::actualLocationChanged(const QUrl & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QMediaRecorder::encoderSettingsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QMediaRecorder::errorOccurred(QMediaRecorder::Error _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2);
}

// SIGNAL 5
void QMediaRecorder::errorChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QMediaRecorder::metaDataChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QMediaRecorder::mediaFormatChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void QMediaRecorder::encodingModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void QMediaRecorder::qualityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void QMediaRecorder::videoResolutionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 10, nullptr);
}

// SIGNAL 11
void QMediaRecorder::videoFrameRateChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void QMediaRecorder::videoBitRateChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 12, nullptr);
}

// SIGNAL 13
void QMediaRecorder::audioBitRateChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void QMediaRecorder::audioChannelCountChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 14, nullptr);
}

// SIGNAL 15
void QMediaRecorder::audioSampleRateChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 15, nullptr);
}

// SIGNAL 16
void QMediaRecorder::autoStopChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 16, nullptr);
}
QT_WARNING_POP
