/****************************************************************************
** Meta object code from reading C++ file 'qaudioroom.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/spatialaudio/qaudioroom.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qaudioroom.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN10QAudioRoomE_t {};
} // unnamed namespace

template <> constexpr inline auto QAudioRoom::qt_create_metaobjectdata<qt_meta_tag_ZN10QAudioRoomE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QAudioRoom",
        "positionChanged",
        "",
        "dimensionsChanged",
        "rotationChanged",
        "wallsChanged",
        "reflectionGainChanged",
        "reverbGainChanged",
        "reverbTimeChanged",
        "reverbBrightnessChanged",
        "position",
        "QVector3D",
        "dimensions",
        "rotation",
        "QQuaternion",
        "reflectionGain",
        "reverbGain",
        "reverbTime",
        "reverbBrightness"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'dimensionsChanged'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'rotationChanged'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'wallsChanged'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reflectionGainChanged'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reverbGainChanged'
        QtMocHelpers::SignalData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reverbTimeChanged'
        QtMocHelpers::SignalData<void()>(8, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reverbBrightnessChanged'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'position'
        QtMocHelpers::PropertyData<QVector3D>(10, 0x80000000 | 11, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 0),
        // property 'dimensions'
        QtMocHelpers::PropertyData<QVector3D>(12, 0x80000000 | 11, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 1),
        // property 'rotation'
        QtMocHelpers::PropertyData<QQuaternion>(13, 0x80000000 | 14, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 2),
        // property 'reflectionGain'
        QtMocHelpers::PropertyData<float>(15, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'reverbGain'
        QtMocHelpers::PropertyData<float>(16, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'reverbTime'
        QtMocHelpers::PropertyData<float>(17, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 6),
        // property 'reverbBrightness'
        QtMocHelpers::PropertyData<float>(18, QMetaType::Float, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 7),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QAudioRoom, qt_meta_tag_ZN10QAudioRoomE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QAudioRoom::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10QAudioRoomE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10QAudioRoomE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10QAudioRoomE_t>.metaTypes,
    nullptr
} };

void QAudioRoom::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QAudioRoom *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->positionChanged(); break;
        case 1: _t->dimensionsChanged(); break;
        case 2: _t->rotationChanged(); break;
        case 3: _t->wallsChanged(); break;
        case 4: _t->reflectionGainChanged(); break;
        case 5: _t->reverbGainChanged(); break;
        case 6: _t->reverbTimeChanged(); break;
        case 7: _t->reverbBrightnessChanged(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::positionChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::dimensionsChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::rotationChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::wallsChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::reflectionGainChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::reverbGainChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::reverbTimeChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (QAudioRoom::*)()>(_a, &QAudioRoom::reverbBrightnessChanged, 7))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QVector3D*>(_v) = _t->position(); break;
        case 1: *reinterpret_cast<QVector3D*>(_v) = _t->dimensions(); break;
        case 2: *reinterpret_cast<QQuaternion*>(_v) = _t->rotation(); break;
        case 3: *reinterpret_cast<float*>(_v) = _t->reflectionGain(); break;
        case 4: *reinterpret_cast<float*>(_v) = _t->reverbGain(); break;
        case 5: *reinterpret_cast<float*>(_v) = _t->reverbTime(); break;
        case 6: *reinterpret_cast<float*>(_v) = _t->reverbBrightness(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setPosition(*reinterpret_cast<QVector3D*>(_v)); break;
        case 1: _t->setDimensions(*reinterpret_cast<QVector3D*>(_v)); break;
        case 2: _t->setRotation(*reinterpret_cast<QQuaternion*>(_v)); break;
        case 3: _t->setReflectionGain(*reinterpret_cast<float*>(_v)); break;
        case 4: _t->setReverbGain(*reinterpret_cast<float*>(_v)); break;
        case 5: _t->setReverbTime(*reinterpret_cast<float*>(_v)); break;
        case 6: _t->setReverbBrightness(*reinterpret_cast<float*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QAudioRoom::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QAudioRoom::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10QAudioRoomE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QAudioRoom::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 8;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 7;
    }
    return _id;
}

// SIGNAL 0
void QAudioRoom::positionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void QAudioRoom::dimensionsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QAudioRoom::rotationChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QAudioRoom::wallsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QAudioRoom::reflectionGainChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QAudioRoom::reverbGainChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void QAudioRoom::reverbTimeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void QAudioRoom::reverbBrightnessChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}
QT_WARNING_POP
