// This file is autogenerated. Changes will be overwritten.
#include "PZONOMFGYT/moc_qaudiobufferinput.cpp"
#include "54G57Q73EM/moc_qpipewire_screencapture_p.cpp"
#include "54G57Q73EM/moc_qpipewire_screencapturehelper_p.cpp"
#include "YR7UWPOU5E/moc_qplatformvideosource_p.cpp"
#include "X7X5W4IIQ4/moc_qpulseaudio_contextmanager_p.cpp"
#include "LQ4XI3EMZE/moc_qvideoframeinput.cpp"
#include "CWRUQVFSPE/moc_qvideoframe.cpp"
#include "CWRUQVFSPE/moc_qvideoframeformat.cpp"
