/****************************************************************************
** Generated QML type registration code
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <QtQml/qqml.h>
#include <QtQml/qqmlmoduleregistration.h>

#if __has_include(<private/qquick3dambientsound_p.h>)
#  include <private/qquick3dambientsound_p.h>
#endif
#if __has_include(<private/qquick3daudioengine_p.h>)
#  include <private/qquick3daudioengine_p.h>
#endif
#if __has_include(<private/qquick3daudiolistener_p.h>)
#  include <private/qquick3daudiolistener_p.h>
#endif
#if __has_include(<private/qquick3daudioroom_p.h>)
#  include <private/qquick3daudioroom_p.h>
#endif
#if __has_include(<private/qquick3dspatialsound_p.h>)
#  include <private/qquick3dspatialsound_p.h>
#endif


#if !defined(QT_STATIC)
#define Q_QMLTYPE_EXPORT Q_DECL_EXPORT
#else
#define Q_QMLTYPE_EXPORT
#endif
Q_QMLTYPE_EXPORT void qml_register_types_QtQuick3D_SpatialAudio()
{
    qmlRegisterModule("QtQuick3D.SpatialAudio", 6, 0);
    QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
    qmlRegisterTypesAndRevisions<QQuick3DAmbientSound>("QtQuick3D.SpatialAudio", 6);
    qmlRegisterEnum<QQuick3DAmbientSound::Loops>("QQuick3DAmbientSound::Loops");
    qmlRegisterTypesAndRevisions<QQuick3DAudioEngine>("QtQuick3D.SpatialAudio", 6);
    qmlRegisterEnum<QQuick3DAudioEngine::OutputMode>("QQuick3DAudioEngine::OutputMode");
    qmlRegisterTypesAndRevisions<QQuick3DAudioListener>("QtQuick3D.SpatialAudio", 6);
    qmlRegisterAnonymousTypesAndRevisions<QQuick3DNode>("QtQuick3D.SpatialAudio", 6);
    qmlRegisterTypesAndRevisions<QQuick3DAudioRoom>("QtQuick3D.SpatialAudio", 6);
    qmlRegisterEnum<QQuick3DAudioRoom::Material>("QQuick3DAudioRoom::Material");
    qmlRegisterTypesAndRevisions<QQuick3DSpatialSound>("QtQuick3D.SpatialAudio", 6);
    qmlRegisterEnum<QQuick3DSpatialSound::DistanceModel>("QQuick3DSpatialSound::DistanceModel");
    qmlRegisterEnum<QQuick3DSpatialSound::Loops>("QQuick3DSpatialSound::Loops");
    QT_WARNING_POP
    qmlRegisterModule("QtQuick3D.SpatialAudio", 6, 12);
}

static const QQmlModuleRegistration qtQuick3DSpatialAudioRegistration("QtQuick3D.SpatialAudio", qml_register_types_QtQuick3D_SpatialAudio);
