/****************************************************************************
** Meta object code from reading C++ file 'qplaybackoptions.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtmultimedia/src/multimedia/playback/qplaybackoptions.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qplaybackoptions.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN16QPlaybackOptionsE_t {};
} // unnamed namespace

template <> constexpr inline auto QPlaybackOptions::qt_create_metaobjectdata<qt_meta_tag_ZN16QPlaybackOptionsE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QPlaybackOptions",
        "RegisterEnumClassesUnscoped",
        "false",
        "networkTimeout",
        "std::chrono::milliseconds",
        "playbackIntent",
        "PlaybackIntent",
        "probeSize",
        "qsizetype",
        "Playback",
        "LowLatencyStreaming"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
        // property 'networkTimeout'
        QtMocHelpers::PropertyData<std::chrono::milliseconds>(3, 0x80000000 | 4, QMC::DefaultPropertyFlags | QMC::Writable | QMC::Resettable | QMC::EnumOrFlag | QMC::StdCppSet | QMC::Final),
        // property 'playbackIntent'
        QtMocHelpers::PropertyData<enum PlaybackIntent>(5, 0x80000000 | 6, QMC::DefaultPropertyFlags | QMC::Writable | QMC::Resettable | QMC::EnumOrFlag | QMC::StdCppSet),
        // property 'probeSize'
        QtMocHelpers::PropertyData<qsizetype>(7, 0x80000000 | 8, QMC::DefaultPropertyFlags | QMC::Writable | QMC::Resettable | QMC::EnumOrFlag | QMC::StdCppSet),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'PlaybackIntent'
        QtMocHelpers::EnumData<enum PlaybackIntent>(6, 6, QMC::EnumIsScoped).add({
            {    9, PlaybackIntent::Playback },
            {   10, PlaybackIntent::LowLatencyStreaming },
        }),
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<QPlaybackOptions, qt_meta_tag_ZN16QPlaybackOptionsE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject QPlaybackOptions::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN16QPlaybackOptionsE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN16QPlaybackOptionsE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN16QPlaybackOptionsE_t>.metaTypes,
    nullptr
} };

void QPlaybackOptions::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = reinterpret_cast<QPlaybackOptions *>(_o);
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<std::chrono::milliseconds*>(_v) = _t->networkTimeout(); break;
        case 1: *reinterpret_cast<enum PlaybackIntent*>(_v) = _t->playbackIntent(); break;
        case 2: *reinterpret_cast<qsizetype*>(_v) = _t->probeSize(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setNetworkTimeout(*reinterpret_cast<std::chrono::milliseconds*>(_v)); break;
        case 1: _t->setPlaybackIntent(*reinterpret_cast<enum PlaybackIntent*>(_v)); break;
        case 2: _t->setProbeSize(*reinterpret_cast<qsizetype*>(_v)); break;
        default: break;
        }
    }
    if (_c == QMetaObject::ResetProperty) {
        switch (_id) {
        case 0: _t->resetNetworkTimeout(); break;
        case 1: _t->resetPlaybackIntent(); break;
        case 2: _t->resetProbeSize(); break;
        default: break;
        }
    }
}
QT_WARNING_POP
